from functools import lru_cache

from django.core.exceptions import PermissionDenied
from django.http import Http404
from django.shortcuts import get_object_or_404, redirect
from django.urls import reverse
from django.utils.text import capfirst
from django.utils.translation import gettext as _
from django.utils.translation import gettext_lazy

from wagtail.admin import messages
from wagtail.admin.panels import (
    ObjectList,
    extract_panel_definitions_from_model_class,
)
from wagtail.admin.ui.side_panels import ChecksSidePanel, PreviewSidePanel
from wagtail.admin.views import generic
from wagtail.admin.views.generic import preview
from wagtail.models import PreviewableMixin, Site

from .forms import SiteSwitchForm
from .models import BaseGenericSetting, BaseSiteSetting
from .registry import registry


def get_model_from_url_params(app_name, model_name):
    """
    retrieve a content type from an app_name / model_name combo.
    Throw Http404 if not a valid setting type
    """
    model = registry.get_by_natural_key(app_name, model_name)
    if model is None:
        raise Http404
    return model


@lru_cache(maxsize=None)
def get_setting_edit_handler(model):
    if hasattr(model, "edit_handler"):
        edit_handler = model.edit_handler
    else:
        if issubclass(model, BaseSiteSetting):
            panels = extract_panel_definitions_from_model_class(model, ["site"])
        elif issubclass(model, BaseGenericSetting):
            panels = extract_panel_definitions_from_model_class(model)
        else:
            raise NotImplementedError

        edit_handler = ObjectList(panels)
    return edit_handler.bind_to_model(model)


def redirect_to_relevant_instance(request, app_name, model_name):
    model = get_model_from_url_params(app_name, model_name)

    if issubclass(model, BaseSiteSetting):
        # Redirect the user to the edit page for the current site
        # (or the current request does not correspond to a site, the first site in the list)
        site = Site.find_for_request(request)
        permission_policy = model.get_permission_policy()
        if not site or not permission_policy.user_has_permission_for_instance(
            request.user, "change", site
        ):
            # Select the first site they can edit
            site = permission_policy.sites_user_has_permission_for(
                request.user, "change"
            ).first()

        if not site:
            if Site.objects.exists():
                # There are sites, but the user has no permission to edit any of them
                raise PermissionDenied
            else:
                # No sites exist, so we can't redirect to a specific site
                messages.error(
                    request,
                    _(
                        "This setting could not be opened because there is no site defined."
                    ),
                )
                return redirect("wagtailadmin_home")

        return redirect(
            "wagtailsettings:edit",
            app_name,
            model_name,
            site.pk,
        )

    elif issubclass(model, BaseGenericSetting):
        return redirect(
            "wagtailsettings:edit",
            app_name,
            model_name,
            model.load(request_or_site=request).id,
        )
    else:
        raise NotImplementedError


class EditView(generic.EditView):
    template_name = "wagtailsettings/edit.html"
    edit_url_name = "wagtailsettings:edit"
    preview_url_name = "wagtailsettings:preview_on_edit"
    error_message = gettext_lazy("The setting could not be saved due to errors.")
    permission_required = "change"

    def setup(self, request, app_name, model_name, *args, **kwargs):
        self.app_name = app_name
        self.model_name = model_name
        self.model = get_model_from_url_params(app_name, model_name)
        self.permission_policy = self.model.get_permission_policy()
        self.pk = kwargs.get(self.pk_url_kwarg)
        super().setup(request, app_name, model_name, *args, **kwargs)

    def get_header_icon(self):
        return registry._model_icons.get(self.model)

    def get_object(self, queryset=None):
        self.site = None
        if issubclass(self.model, BaseSiteSetting):
            self.site = get_object_or_404(Site, pk=self.pk)

            if not self.permission_policy.user_has_permission_for_instance(
                self.request.user, "change", self.site
            ):
                raise PermissionDenied

            return self.model.for_site(self.site)
        else:
            return get_object_or_404(self.model, pk=self.pk)

    def get_panel(self):
        return get_setting_edit_handler(self.model)

    def get_edit_url(self):
        return reverse(
            self.edit_url_name,
            args=(self.app_name, self.model_name, self.pk),
        )

    def get_success_buttons(self):
        return None

    def get_page_subtitle(self):
        return capfirst(self.model._meta.verbose_name)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        site_switcher = None
        site_for_header = (
            None  # shown when multiple sites exist, but only one is editable
        )

        if issubclass(self.model, BaseSiteSetting):
            sites = self.permission_policy.sites_user_has_permission_for(
                self.request.user, "change"
            )
            if len(sites) > 1:
                site_switcher = SiteSwitchForm(self.site, self.model, sites=sites)
                context["media"] += site_switcher.media
            elif Site.objects.count() > 1:
                site_for_header = self.site

        context["site_switcher"] = site_switcher
        context["site_for_header"] = site_for_header
        return context

    def get_success_url(self):
        return self.get_edit_url()

    def get_success_message(self):
        return capfirst(
            _("%(setting_type)s updated.")
            % {"setting_type": self.model._meta.verbose_name}
        )

    def get_preview_url(self):
        if isinstance(self.object, PreviewableMixin) and self.object.is_previewable():
            return reverse(
                self.preview_url_name,
                args=(self.app_name, self.model_name, self.pk),
            )

    def get_side_panels(self):
        side_panels = super().get_side_panels()
        if preview_url := self.get_preview_url():
            side_panels.append(
                PreviewSidePanel(self.object, self.request, preview_url=preview_url)
            )
            side_panels.append(ChecksSidePanel(self.object, self.request))
        return side_panels


class PreviewOnEdit(preview.PreviewOnEdit):
    def setup(self, request, app_name, model_name, *args, **kwargs):
        self.app_name = app_name
        self.model_name = model_name
        self.model = get_model_from_url_params(app_name, model_name)
        self.permission_policy = self.model.get_permission_policy()
        self.pk = kwargs.get("pk")
        super().setup(request, app_name, model_name, *args, **kwargs)

    def get_object(self, queryset=None):
        self.site = None
        if issubclass(self.model, BaseSiteSetting):
            self.site = get_object_or_404(Site, pk=self.pk)
            obj = self.model.for_site(self.site)
        else:
            obj = get_object_or_404(self.model, pk=self.pk)

        if not self.user_has_permission_for_instance(self.permission_required, obj):
            raise PermissionDenied

        return obj

    def get_extra_request_attrs(self):
        attrs = super().get_extra_request_attrs()
        # Inject current instance into request so the context processor uses it
        attrs[self.object.get_cache_attr_name()] = self.object
        return attrs



from typing import Any

from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import PermissionDenied
from django.http import Http404
from django.urls import reverse
from django.utils.functional import cached_property, classproperty
from django.utils.text import capfirst
from django.utils.translation import gettext
from django.utils.translation import gettext_lazy as _

from wagtail.admin.views import generic
from wagtail.admin.views.generic.base import BaseListingView
from wagtail.admin.views.generic.permissions import PermissionCheckedMixin
from wagtail.admin.views.pages.listing import PageFilterSet, PageListingMixin
from wagtail.admin.views.pages.utils import (
    GenericPageBreadcrumbsMixin,
)
from wagtail.models import Page
from wagtail.permissions import page_permission_policy


class ContentTypeUseView(PageListingMixin, PermissionCheckedMixin, BaseListingView):
    permission_policy = page_permission_policy
    any_permission_required = {
        "add",
        "change",
        "publish",
        "bulk_delete",
        "lock",
        "unlock",
    }
    index_url_name = "wagtailadmin_pages:type_use"
    index_results_url_name = "wagtailadmin_pages:type_use_results"
    page_title = _("Pages using")
    header_icon = "doc-empty-inverse"
    paginate_by = 50
    filterset_class = PageFilterSet

    @classproperty
    def columns(cls):
        return [col for col in PageListingMixin.columns if col.name != "type"]

    def get(self, request, *, content_type_app_name, content_type_model_name):
        try:
            content_type = ContentType.objects.get_by_natural_key(
                content_type_app_name, content_type_model_name
            )
        except ContentType.DoesNotExist:
            raise Http404

        self.page_class = content_type.model_class()

        # page_class must be a Page type and not some other random model
        if not issubclass(self.page_class, Page):
            raise Http404

        return super().get(request)

    def get_page_subtitle(self):
        return self.page_class.get_verbose_name()

    @cached_property
    def verbose_name_plural(self):
        return self.page_class._meta.verbose_name_plural

    def get_base_queryset(self):
        queryset = self.page_class._default_manager.all().filter(
            pk__in=self.permission_policy.explorable_instances(
                self.request.user
            ).values_list("pk", flat=True)
        )
        return self.annotate_queryset(queryset)

    def get_index_url(self):
        return reverse(
            self.index_url_name,
            args=[
                self.kwargs["content_type_app_name"],
                self.kwargs["content_type_model_name"],
            ],
        )

    def get_index_results_url(self):
        return reverse(
            self.index_results_url_name,
            args=[
                self.kwargs["content_type_app_name"],
                self.kwargs["content_type_model_name"],
            ],
        )

    def get_breadcrumbs_items(self):
        return self.breadcrumbs_items + [
            # Initially, this page was only accessible via the "choose page type"
            # step of the page creation flow. However, that view might be skipped
            # if there's only one valid page type. Since then, we've introduced
            # the "page types usage" report, which always provides a link to this
            # view for all page types. Use that as the parent link.
            {
                "url": reverse("wagtailadmin_reports:page_types_usage"),
                "label": gettext("Page types usage"),
            },
            {
                "url": self.get_index_url(),
                "label": capfirst(self.page_class._meta.verbose_name_plural),
            },
        ]

    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        context = super().get_context_data(**kwargs)
        context["page_class"] = self.page_class
        return context


class UsageView(GenericPageBreadcrumbsMixin, generic.UsageView):
    model = Page
    pk_url_kwarg = "page_id"
    header_icon = "doc-empty-inverse"
    usage_url_name = "wagtailadmin_pages:usage"
    edit_url_name = "wagtailadmin_pages:edit"

    def dispatch(self, request, *args, **kwargs):
        if not self.object.permissions_for_user(request.user).can_edit():
            raise PermissionDenied
        return super().dispatch(request, *args, **kwargs)



from datetime import date
from unittest import mock

from django.test import TestCase, override_settings

from wagtail.models import Page
from wagtail.search import index
from wagtail.test.search import models
from wagtail.test.testapp.models import AdvertWithCustomUUIDPrimaryKey, SimplePage
from wagtail.test.utils import WagtailTestUtils


class TestGetIndexedInstance(TestCase):
    fixtures = ["search"]

    def test_gets_instance(self):
        obj = models.Author.objects.get(id=1)

        # Should just return the object
        indexed_instance = index.get_indexed_instance(obj)
        self.assertEqual(indexed_instance, obj)

    def test_gets_specific_class(self):
        obj = models.Novel.objects.get(id=1)

        # Running the command with the parent class should find the specific class again
        indexed_instance = index.get_indexed_instance(obj.book_ptr)
        self.assertEqual(indexed_instance, obj)

    def test_blocks_not_in_indexed_objects(self):
        obj = models.Novel(
            title="Don't index me!",
            publication_date=date(2017, 10, 18),
            number_of_pages=100,
        )
        obj.save()

        # We've told it not to index anything with the title "Don't index me"
        # get_indexed_instance should return None
        indexed_instance = index.get_indexed_instance(obj.book_ptr)
        self.assertIsNone(indexed_instance)


@mock.patch("wagtail.search.tests.DummySearchBackend", create=True)
@override_settings(
    WAGTAILSEARCH_BACKENDS={
        "default": {"BACKEND": "wagtail.search.tests.DummySearchBackend"}
    }
)
class TestInsertOrUpdateObject(WagtailTestUtils, TestCase):
    def test_inserts_object(self, backend):
        obj = models.Book.objects.create(
            title="Test", publication_date=date(2017, 10, 18), number_of_pages=100
        )
        backend().reset_mock()

        index.insert_or_update_object(obj)

        backend().add.assert_called_with(obj)

    def test_doesnt_insert_unsaved_object(self, backend):
        obj = models.Book(
            title="Test", publication_date=date(2017, 10, 18), number_of_pages=100
        )
        backend().reset_mock()

        index.insert_or_update_object(obj)

        self.assertFalse(backend().add.mock_calls)

    def test_converts_to_specific_page(self, backend):
        root_page = Page.objects.get(id=1)
        page = root_page.add_child(
            instance=SimplePage(title="test", slug="test", content="test")
        )

        # Convert page into a generic "Page" object and add it into the index
        unspecific_page = page.page_ptr

        backend().reset_mock()

        index.insert_or_update_object(unspecific_page)

        # It should be automatically converted back to the specific version
        backend().add.assert_called_with(page)

    def test_catches_index_error(self, backend):
        obj = models.Book.objects.create(
            title="Test", publication_date=date(2017, 10, 18), number_of_pages=100
        )

        backend().add.side_effect = ValueError("Test")
        backend().reset_mock()

        with self.assertLogs("wagtail.search.index", level="ERROR") as cm:
            index.insert_or_update_object(obj)

        self.assertEqual(len(cm.output), 1)
        self.assertIn(
            "Exception raised while adding <Book: Test> into the 'default' search backend",
            cm.output[0],
        )
        self.assertIn("Traceback (most recent call last):", cm.output[0])
        self.assertIn("ValueError: Test", cm.output[0])


@mock.patch("wagtail.search.tests.DummySearchBackend", create=True)
@override_settings(
    WAGTAILSEARCH_BACKENDS={
        "default": {"BACKEND": "wagtail.search.tests.DummySearchBackend"}
    }
)
class TestRemoveObject(WagtailTestUtils, TestCase):
    def test_removes_object(self, backend):
        obj = models.Book.objects.create(
            title="Test", publication_date=date(2017, 10, 18), number_of_pages=100
        )
        backend().reset_mock()

        index.remove_object(obj)

        backend().delete.assert_called_with(obj)

    def test_removes_unsaved_object(self, backend):
        obj = models.Book(
            title="Test", publication_date=date(2017, 10, 18), number_of_pages=100
        )
        backend().reset_mock()

        index.remove_object(obj)

        backend().delete.assert_called_with(obj)

    def test_catches_index_error(self, backend):
        obj = models.Book.objects.create(
            title="Test", publication_date=date(2017, 10, 18), number_of_pages=100
        )
        backend().reset_mock()

        backend().delete.side_effect = ValueError("Test")

        with self.assertLogs("wagtail.search.index", level="ERROR") as cm:
            index.remove_object(obj)

        self.assertEqual(len(cm.output), 1)
        self.assertIn(
            "Exception raised while deleting <Book: Test> from the 'default' search backend",
            cm.output[0],
        )
        self.assertIn("Traceback (most recent call last):", cm.output[0])
        self.assertIn("ValueError: Test", cm.output[0])


@mock.patch("wagtail.search.tests.DummySearchBackend", create=True)
@override_settings(
    WAGTAILSEARCH_BACKENDS={
        "default": {"BACKEND": "wagtail.search.tests.DummySearchBackend"}
    }
)
class TestSignalHandlers(WagtailTestUtils, TestCase):
    def test_index_on_create(self, backend):
        backend().reset_mock()
        with self.captureOnCommitCallbacks(execute=True):
            obj = models.Book.objects.create(
                title="Test", publication_date=date(2017, 10, 18), number_of_pages=100
            )
        backend().add.assert_called_with(obj)

    def test_index_on_create_with_uuid_primary_key(self, backend):
        backend().reset_mock()
        with self.captureOnCommitCallbacks(execute=True):
            obj = AdvertWithCustomUUIDPrimaryKey.objects.create(text="Test")
        backend().add.assert_called_with(obj)

    def test_index_on_update(self, backend):
        obj = models.Book.objects.create(
            title="Test", publication_date=date(2017, 10, 18), number_of_pages=100
        )

        backend().reset_mock()
        obj.title = "Updated test"
        with self.captureOnCommitCallbacks(execute=True):
            obj.save()

        self.assertEqual(backend().add.call_count, 1)
        indexed_object = backend().add.call_args[0][0]
        self.assertEqual(indexed_object.title, "Updated test")

    def test_index_on_update_with_uuid_primary_key(self, backend):
        obj = AdvertWithCustomUUIDPrimaryKey.objects.create(text="Test")

        backend().reset_mock()
        obj.text = "Updated test"
        with self.captureOnCommitCallbacks(execute=True):
            obj.save()

        self.assertEqual(backend().add.call_count, 1)
        indexed_object = backend().add.call_args[0][0]
        self.assertEqual(indexed_object.text, "Updated test")

    def test_index_on_delete(self, backend):
        obj = models.Book.objects.create(
            title="Test", publication_date=date(2017, 10, 18), number_of_pages=100
        )

        backend().reset_mock()
        with self.captureOnCommitCallbacks(execute=True):
            obj.delete()
        backend().delete.assert_called_with(obj)

    def test_index_on_delete_with_uuid_primary_key(self, backend):
        obj = AdvertWithCustomUUIDPrimaryKey.objects.create(text="Test")

        backend().reset_mock()
        with self.captureOnCommitCallbacks(execute=True):
            obj.delete()
        backend().delete.assert_called_with(obj)

    def test_do_not_index_fields_omitted_from_update_fields(self, backend):
        obj = models.Book.objects.create(
            title="Test", publication_date=date(2017, 10, 18), number_of_pages=100
        )

        backend().reset_mock()
        obj.title = "Updated test"
        obj.publication_date = date(2001, 10, 19)
        with self.captureOnCommitCallbacks(execute=True):
            obj.save(update_fields=["title"])

        self.assertEqual(backend().add.call_count, 1)
        indexed_object = backend().add.call_args[0][0]
        self.assertEqual(indexed_object.title, "Updated test")
        self.assertEqual(indexed_object.publication_date, date(2017, 10, 18))


@mock.patch("wagtail.search.tests.DummySearchBackend", create=True)
@override_settings(
    WAGTAILSEARCH_BACKENDS={
        "default": {"BACKEND": "wagtail.search.tests.DummySearchBackend"}
    }
)
class TestSignalHandlersSearchDisabled(TestCase, WagtailTestUtils):
    def test_index_on_create_and_update(self, backend):
        obj = models.UnindexedBook.objects.create(
            title="Test", publication_date=date(2017, 10, 18), number_of_pages=100
        )

        self.assertEqual(backend().add.call_count, 0)
        self.assertIsNone(backend().add.call_args)

        backend().reset_mock()
        obj.title = "Updated test"
        obj.save()

        self.assertEqual(backend().add.call_count, 0)
        self.assertIsNone(backend().add.call_args)



from django.core.exceptions import PermissionDenied
from django.db import transaction

from wagtail.coreutils import find_available_slug
from wagtail.models.copying import _copy
from wagtail.signals import copy_for_translation_done


class ParentNotTranslatedError(Exception):
    """
    Raised when a call to Page.copy_for_translation is made but the
    parent page is not translated and copy_parents is False.
    """

    pass


class CopyForTranslationPermissionError(PermissionDenied):
    """
    Raised when the object translation copy cannot be performed due to insufficient permissions.
    """

    pass


class CopyPageForTranslationPermissionError(CopyForTranslationPermissionError):
    pass


class CopyPageForTranslationAction:
    """
    Creates a copy of this page in the specified locale.

    The new page will be created in draft as a child of this page's translated
    parent.

    For example, if you are translating a blog post from English into French,
    this method will look for the French version of the blog index and create
    the French translation of the blog post under that.

    If this page's parent is not translated into the locale, then a ``ParentNotTranslatedError``
    is raised. You can circumvent this error by passing ``copy_parents=True`` which
    copies any parents that are not translated yet.

    The ``exclude_fields`` parameter can be used to set any fields to a blank value
    in the copy.

    Note that this method calls the ``.copy()`` method internally so any fields that
    are excluded in ``.exclude_fields_in_copy`` will be excluded from the translation.
    """

    def __init__(
        self,
        page,
        locale,
        copy_parents=False,
        alias=False,
        exclude_fields=None,
        user=None,
        include_subtree=False,
    ):
        self.page = page
        self.locale = locale
        self.copy_parents = copy_parents
        self.alias = alias
        self.exclude_fields = exclude_fields
        self.user = user
        self.include_subtree = include_subtree

    def check(self, skip_permission_checks=False):
        # Permission checks
        if (
            self.user
            and not skip_permission_checks
            and not self.user.has_perms(["simple_translation.submit_translation"])
        ):
            raise CopyPageForTranslationPermissionError(
                "You do not have permission to submit a translation for this page."
            )

    def walk(self, current_page):
        for child_page in current_page.get_children():
            translated_page = self._copy_for_translation(
                child_page
                if child_page.live
                else child_page.get_latest_revision_as_object(),
                self.locale,
                self.copy_parents,
                self.alias,
                self.exclude_fields,
            )

            copy_for_translation_done.send(
                sender=self.__class__,
                source_obj=child_page.specific,
                target_obj=translated_page,
            )

            self.walk(child_page)

    @transaction.atomic
    def _copy_for_translation(self, page, locale, copy_parents, alias, exclude_fields):
        # Find the translated version of the parent page to create the new page under
        parent = page.get_parent().specific
        slug = page.slug

        if not parent.is_root():
            try:
                translated_parent = parent.get_translation(locale)
            except parent.__class__.DoesNotExist:
                if not copy_parents:
                    raise ParentNotTranslatedError("Parent page is not translated.")

                translated_parent = parent.copy_for_translation(
                    locale, copy_parents=True, alias=True
                )
        else:
            # Don't duplicate the root page for translation. Create new locale as a sibling
            translated_parent = parent

            # Append language code to slug as the new page
            # will be created in the same section as the existing one
            slug += "-" + locale.language_code

        # Find available slug for new page
        slug = find_available_slug(translated_parent, slug)

        if alias:
            return page.create_alias(
                parent=translated_parent,
                update_slug=slug,
                update_locale=locale,
                reset_translation_key=False,
            )

        else:
            # Update locale on translatable child objects as well
            def process_child_object(
                original_page, page_copy, child_relation, child_object
            ):
                from wagtail.models import TranslatableMixin

                if isinstance(child_object, TranslatableMixin):
                    child_object.locale = locale

            return page.copy(
                to=translated_parent,
                update_attrs={
                    "locale": locale,
                    "slug": slug,
                },
                copy_revisions=False,
                keep_live=False,
                reset_translation_key=False,
                process_child_object=process_child_object,
                exclude_fields=exclude_fields,
                log_action="wagtail.copy_for_translation",
            )

    def execute(self, skip_permission_checks=False):
        self.check(skip_permission_checks=skip_permission_checks)

        translated_page = self._copy_for_translation(
            self.page if self.page.live else self.page.get_latest_revision_as_object(),
            self.locale,
            self.copy_parents,
            self.alias,
            self.exclude_fields,
        )

        copy_for_translation_done.send(
            sender=self.__class__,
            source_obj=self.page,
            target_obj=translated_page,
        )

        if self.include_subtree:
            self.walk(self.page)

        return translated_page


class CopyForTranslationAction:
    """
    Creates a copy of this object in the specified locale.

    The ``exclude_fields`` parameter can be used to set any fields to a blank value
    in the copy.
    """

    def __init__(
        self,
        object,
        locale,
        exclude_fields=None,
        user=None,
    ):
        self.object = object
        self.locale = locale
        self.exclude_fields = exclude_fields
        self.user = user

    def check(self, skip_permission_checks=False):
        # Permission checks
        if (
            self.user
            and not skip_permission_checks
            and not self.user.has_perms(["simple_translation.submit_translation"])
        ):
            raise CopyForTranslationPermissionError(
                "You do not have permission to submit a translation for this object."
            )

    @transaction.atomic
    def _copy_for_translation(self, object, locale, exclude_fields=None):
        from wagtail.models import DraftStateMixin, TranslatableMixin

        # Make sure the copy includes the latest changes, including draft
        if isinstance(object, DraftStateMixin):
            object = object.get_latest_revision_as_object()

        exclude_fields = (
            getattr(object, "default_exclude_fields_in_copy", [])
            + getattr(object, "exclude_fields_in_copy", [])
            + (exclude_fields or [])
        )
        translated, child_object_map = _copy(object, exclude_fields=exclude_fields)
        translated.locale = locale

        # Update locale on any translatable child objects as well
        # Note: If this is not a subclass of ClusterableModel, child_object_map will always be '{}'
        for (_child_relation, _old_pk), child_object in child_object_map.items():
            if isinstance(child_object, TranslatableMixin):
                child_object.locale = locale

        return translated

    def execute(self, skip_permission_checks=False):
        self.check(skip_permission_checks=skip_permission_checks)

        translated_object = self._copy_for_translation(
            self.object, self.locale, self.exclude_fields
        )

        copy_for_translation_done.send(
            sender=self.__class__,
            source_obj=self.object,
            target_obj=translated_object,
        )

        return translated_object



from factory import declarations
from factory.base import FactoryOptions, OptionDefault


class BlockFactoryOptions(FactoryOptions):
    def _build_default_options(self):
        options = super()._build_default_options()
        options.append(OptionDefault("block_def", None))
        return options

    def get_meta_dict(self):
        return {
            "model": self.model,
            "block_def": self.block_def,
            "abstract": self.abstract,
            "strategy": self.strategy,
            "inline_args": self.inline_args,
            "exclude": self.exclude,
            "rename": self.rename,
        }

    def to_meta_class(self):
        """
        Create a new Meta class from this instance's options, suitable for
        inclusion on a factory subclass
        """
        return type("Meta", (), self.get_meta_dict())


class StreamBlockFactoryOptions(BlockFactoryOptions):
    def prepare_arguments(self, attributes):
        # Like the base implementation, but ignore args as they are not relevant
        # for instantiating StreamValues.

        def get_block_name(key):
            # Keys at this point will be like <index>.<block_name>
            return key.split(".")[1]

        kwargs = dict(attributes)
        # 1. Extension points
        kwargs = self.factory._adjust_kwargs(**kwargs)

        # 2. Remove hidden objects
        filtered_kwargs = {}
        for k, v in kwargs.items():
            block_name = get_block_name(k)
            if (
                block_name not in self.exclude
                and block_name not in self.parameters
                and v is not declarations.SKIP
            ):
                filtered_kwargs[k] = v

        return (), filtered_kwargs

    def get_block_definition(self):
        if self.block_def is not None:
            return self.block_def
        elif self.model is not None:
            return self.model()



from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class WagtailSitesAppConfig(AppConfig):
    name = "wagtail.sites"
    label = "wagtailsites"
    verbose_name = _("Wagtail sites")



from wagtail import hooks


class FeatureRegistry:
    """
    A central store of information about optional features that can be enabled in rich text
    editors by passing a ``features`` list to the RichTextField, such as how to
    whitelist / convert HTML tags, and how to enable the feature on various editors.

    This information may come from diverse sources - for example, wagtailimages might define
    an 'images' feature and a Draftail plugin for it, while a third-party module might
    define a TinyMCE plugin for the same feature. The information is therefore collected into
    this registry via the 'register_rich_text_features' hook.
    """

    def __init__(self):
        # Has the register_rich_text_features hook been run for this registry?
        self.has_scanned_for_features = False

        # a dict of dicts, one for each editor (draftail.js, TinyMCE etc); each dict is a mapping
        # of feature names to 'plugin' objects that define how to implement that feature
        # (e.g. paths to JS files to import). The API of that plugin object is not defined
        # here, and is specific to each editor.
        self.plugins_by_editor = {}

        # a list of feature names that will be applied on rich text areas that do not specify
        # an explicit `feature` list.
        self.default_features = []

        # a mapping of linktype names to rewriter functions for converting database representations
        # of links (e.g. <a linktype="page" id="123">) into front-end HTML. Each rewriter function
        # takes a dict of attributes, and returns the rewritten opening tag as a string
        self.link_types = {}

        # a mapping of embedtype names to rewriter functions for converting database representations
        # of embedded content (e.g. <embed embedtype="image" id="123" format="left" alt="foo">)
        # into front-end HTML. Each rewriter function takes a dict of attributes, and returns an
        # HTML fragment to replace it with
        self.embed_types = {}

        # a dict of dicts, one for each converter backend (editorhtml, contentstate etc);
        # each dict is a mapping of feature names to 'rule' objects that define how to convert
        # that feature's elements between editor representation and database representation
        # (e.g. elements to whitelist, functions for transferring attributes).
        # The API of that rule object is not defined here, and is specific to each converter backend.
        self.converter_rules_by_converter = {}

    def get_default_features(self):
        if not self.has_scanned_for_features:
            self._scan_for_features()

        return self.default_features

    def _scan_for_features(self):
        for fn in hooks.get_hooks("register_rich_text_features"):
            fn(self)
        self.has_scanned_for_features = True

    def register_editor_plugin(self, editor_name, feature_name, plugin):
        plugins = self.plugins_by_editor.setdefault(editor_name, {})
        plugins[feature_name] = plugin

    def get_editor_plugin(self, editor_name, feature_name):
        if not self.has_scanned_for_features:
            self._scan_for_features()

        try:
            return self.plugins_by_editor[editor_name][feature_name]
        except KeyError:
            return None

    def register_link_type(self, handler):
        self.link_types[handler.identifier] = handler

    def get_link_types(self):
        if not self.has_scanned_for_features:
            self._scan_for_features()
        return self.link_types

    def register_embed_type(self, handler):
        self.embed_types[handler.identifier] = handler

    def get_embed_types(self):
        if not self.has_scanned_for_features:
            self._scan_for_features()
        return self.embed_types

    def register_converter_rule(self, converter_name, feature_name, rule):
        rules = self.converter_rules_by_converter.setdefault(converter_name, {})
        rules[feature_name] = rule

    def get_converter_rule(self, converter_name, feature_name):
        if not self.has_scanned_for_features:
            self._scan_for_features()

        try:
            return self.converter_rules_by_converter[converter_name][feature_name]
        except KeyError:
            return None

    @staticmethod
    def function_as_entity_handler(identifier, fn):
        """Supports legacy registering of entity handlers as functions."""
        return type(
            "EntityHandlerRegisteredAsFunction",
            (object,),
            {
                "identifier": identifier,
                "expand_db_attributes": staticmethod(fn),
            },
        )



from wagtail.models import Page, get_translatable_models


def get_locale_usage(locale):
    """
    Returns the number of pages and other objects that use a locale
    """
    num_pages = Page.objects.filter(locale=locale).exclude(depth=1).count()

    num_others = 0

    for model in get_translatable_models():
        if model is Page:
            continue

        num_others += model.objects.filter(locale=locale).count()

    return num_pages, num_others



from django.contrib.auth.models import Permission
from django.test import TestCase
from django.urls import reverse

from wagtail.admin.staticfiles import versioned_static
from wagtail.models import Site
from wagtail.test.testapp.models import (
    PreviewableGenericSetting,
    PreviewableSiteSetting,
)
from wagtail.test.utils import WagtailTestUtils


class TestGenericSiteSettingPreview(WagtailTestUtils, TestCase):
    model = PreviewableGenericSetting
    other_model = PreviewableSiteSetting

    def setUp(self):
        self.user = self.login()
        site = Site.objects.get(is_default_site=True)

        PreviewableGenericSetting.objects.create(
            text="An initial previewable generic setting"
        )
        PreviewableSiteSetting.objects.create(
            text="An initial previewable site setting",
            site=site,
        )

        self.setting = self.model.objects.first()
        self.app_label = self.model._meta.app_label
        self.model_name = self.model._meta.model_name
        self.verbose_name = self.model._meta.verbose_name
        url_pk = getattr(self.setting, "site_id", self.setting.pk)
        self.preview_on_edit_url = reverse(
            "wagtailsettings:preview_on_edit",
            args=(self.app_label, self.model_name, url_pk),
        )
        self.session_key_prefix = f"wagtail-preview-{self.app_label}-{self.model_name}"
        self.edit_session_key = f"{self.session_key_prefix}-{self.setting.pk}"

        self.post_data = {
            "text": f"An edited {self.verbose_name}",
        }

    def test_preview_on_edit_with_valid_then_invalid_data(self):
        response = self.client.post(self.preview_on_edit_url, self.post_data)

        # Check the JSON response
        self.assertEqual(response.status_code, 200)
        self.assertJSONEqual(
            response.content.decode(),
            {"is_valid": True, "is_available": True},
        )

        # Send an invalid update request
        response = self.client.post(self.preview_on_edit_url, {})
        self.assertEqual(response.status_code, 200)
        self.assertJSONEqual(
            response.content.decode(),
            {"is_valid": False, "is_available": True},
        )

        # Check the user can still see the preview with the last valid data
        self.assertIn(self.edit_session_key, self.client.session)

        response = self.client.get(self.preview_on_edit_url)

        # Check the HTML response
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, "tests/previewable_setting.html")
        soup = self.get_soup(response.content)
        # This value is accessed using the default `object` context variable
        current = soup.select_one("#current-object")
        self.assertIsNotNone(current)
        self.assertEqual(current.text.strip(), f"An edited {self.verbose_name}")
        # This value is accessed using the context processor, and it's the
        # setting being previewed
        injected = soup.select_one(f"#{self.model_name}")
        self.assertIsNotNone(injected)
        self.assertEqual(injected.text.strip(), f"An edited {self.verbose_name}")
        # The other setting is accessed using the context processor and it
        # should not be affected and still have the initial value
        other_setting = soup.select_one(f"#{self.other_model._meta.model_name}")
        self.assertIsNotNone(other_setting)
        self.assertEqual(
            other_setting.text.strip(),
            f"An initial {self.other_model._meta.verbose_name}",
        )

    def test_preview_on_edit_clear_preview_data(self):
        # Set a fake preview session data for the setting
        self.client.session[self.edit_session_key] = "test data"

        response = self.client.delete(self.preview_on_edit_url)
        self.assertEqual(response.status_code, 200)
        self.assertJSONEqual(
            response.content.decode(),
            {"success": True},
        )

        # The data should no longer exist in the session
        self.assertNotIn(self.edit_session_key, self.client.session)

        response = self.client.get(self.preview_on_edit_url)

        # The preview should be unavailable
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(response, "wagtailadmin/generic/preview_error.html")
        self.assertContains(
            response,
            "<title>Preview not available - Wagtail</title>",
            html=True,
        )
        self.assertContains(
            response,
            '<h1 class="preview-error__title">Preview not available</h1>',
            html=True,
        )
        self.assertNotContains(response, versioned_static("wagtailadmin/js/icons.js"))

    def test_preview_on_edit_without_permission(self):
        # Remove privileges from user
        self.user.is_superuser = False
        self.user.user_permissions.add(
            Permission.objects.get(
                content_type__app_label="wagtailadmin", codename="access_admin"
            )
        )
        self.user.save()

        response = self.client.post(self.preview_on_edit_url, self.post_data)
        self.assertEqual(response.status_code, 302)
        self.assertRedirects(response, reverse("wagtailadmin_home"))

    def test_preview_on_edit_get_without_permission(self):
        # Remove privileges from user
        self.user.is_superuser = False
        self.user.user_permissions.add(
            Permission.objects.get(
                content_type__app_label="wagtailadmin", codename="access_admin"
            )
        )
        self.user.save()

        response = self.client.get(self.preview_on_edit_url)
        self.assertEqual(response.status_code, 302)
        self.assertRedirects(response, reverse("wagtailadmin_home"))


class TestSiteSettingPreview(TestGenericSiteSettingPreview):
    model = PreviewableSiteSetting
    other_model = PreviewableGenericSetting


class TestEnablePreviewGenericSetting(WagtailTestUtils, TestCase):
    model = PreviewableGenericSetting

    def setUp(self):
        self.user = self.login()
        site = Site.objects.get(is_default_site=True)
        PreviewableGenericSetting.objects.create(
            text="An initial previewable generic setting"
        )
        PreviewableSiteSetting.objects.create(
            text="An initial previewable site setting",
            site=site,
        )
        self.setting = self.model.objects.first()
        self.url_pk = getattr(self.setting, "site_id", self.setting.pk)

    def get_url(self, name):
        return reverse(
            f"wagtailsettings:{name}",
            args=("tests", self.setting._meta.model_name, self.url_pk),
        )

    def test_show_preview_panel_on_edit(self):
        edit_url = self.get_url("edit")
        preview_url = self.get_url("preview_on_edit")
        new_tab_url = preview_url + "?mode="
        response = self.client.get(edit_url)

        self.assertEqual(response.status_code, 200)

        # Should have the preview side panel toggle button
        soup = self.get_soup(response.content)
        self.assertIsNotNone(soup.select_one('[data-side-panel="preview"]'))
        toggle_button = soup.find("button", {"data-side-panel-toggle": "preview"})
        self.assertIsNotNone(toggle_button)
        self.assertEqual("w-tooltip w-kbd", toggle_button["data-controller"])
        self.assertEqual("mod+p", toggle_button["data-w-kbd-key-value"])

        # Should have the checks side panel
        soup = self.get_soup(response.content)
        self.assertIsNotNone(soup.select_one('[data-side-panel="checks"]'))
        toggle_button = soup.find("button", {"data-side-panel-toggle": "checks"})
        self.assertIsNotNone(toggle_button)

        # Should set the preview URL value on the controller
        controller = soup.select_one('[data-controller="w-preview"]')
        self.assertIsNotNone(controller)
        self.assertEqual(controller.get("data-w-preview-url-value"), preview_url)

        # Should show the iframe
        iframe = controller.select_one("#w-preview-iframe")
        self.assertIsNotNone(iframe)
        self.assertEqual(iframe.get("data-w-preview-target"), "iframe")

        # Should show the new tab button with the default mode set
        new_tab_button = controller.select_one('a[data-w-preview-target="newTab"]')
        self.assertIsNotNone(new_tab_button)
        self.assertEqual(new_tab_button["href"], new_tab_url)
        self.assertEqual(new_tab_button["target"], "_blank")

        # Should not show the preview mode selection
        mode_select = controller.select_one('[data-w-preview-target="mode"]')
        self.assertIsNone(mode_select)


class TestEnablePreviewSiteSetting(TestEnablePreviewGenericSetting):
    model = PreviewableSiteSetting



from django.test import TestCase

from wagtail.contrib.redirects.forms import ConfirmImportForm, ImportForm


class TestImportForm(TestCase):
    def test_file_input_generates_proper_accept(self):
        form = ImportForm(["csv", "tsv"])
        self.assertIn('accept=".csv,.tsv"', form.as_table())


class TestConfirmImportForm(TestCase):
    def test_choices_get_appended_with_intro_label_if_multiple(self):
        form = ConfirmImportForm(headers=[(0, "From"), (1, "To")])
        first_choice = form.fields["from_index"].choices[0]
        self.assertEqual(first_choice[0], "")
        self.assertEqual(first_choice[1], "---")

    def test_choices_does_not_get_generated_label_if_single_choice(self):
        form = ConfirmImportForm(headers=[(1, "Hi")])
        first_choice = form.fields["from_index"].choices[0]
        self.assertNotEqual(first_choice[0], "")
        self.assertNotEqual(first_choice[1], "---")



from functools import wraps

from django.conf import settings
from django.core.exceptions import PermissionDenied
from django.shortcuts import redirect
from django.urls import reverse
from django.utils.translation import gettext as _

from wagtail.admin import messages
from wagtail.admin.localization import get_localized_response
from wagtail.log_actions import LogContext
from wagtail.permissions import page_permission_policy


def permission_denied(request):
    """Return a standard 'permission denied' response"""
    if request.headers.get("x-requested-with") == "XMLHttpRequest":
        raise PermissionDenied

    from wagtail.admin import messages

    messages.error(request, _("Sorry, you do not have permission to access this area."))
    return redirect("wagtailadmin_home")


def user_passes_test(test):
    """
    Given a test function that takes a user object and returns a boolean,
    return a view decorator that denies access to the user if the test returns false.
    """

    def decorator(view_func):
        # decorator takes the view function, and returns the view wrapped in
        # a permission check

        @wraps(view_func)
        def wrapped_view_func(request, *args, **kwargs):
            if test(request.user):
                # permission check succeeds; run the view function as normal
                return view_func(request, *args, **kwargs)
            else:
                # permission check failed
                return permission_denied(request)

        return wrapped_view_func

    return decorator


def permission_required(permission_name):
    """
    Replacement for django.contrib.auth.decorators.permission_required which returns a
    more meaningful 'permission denied' response than just redirecting to the login page.
    (The latter doesn't work anyway because Wagtail doesn't define LOGIN_URL...)
    """

    def test(user):
        return user.has_perm(permission_name)

    # user_passes_test constructs a decorator function specific to the above test function
    return user_passes_test(test)


def any_permission_required(*perms):
    """
    Decorator that accepts a list of permission names, and allows the user
    to pass if they have *any* of the permissions in the list
    """

    def test(user):
        for perm in perms:
            if user.has_perm(perm):
                return True

        return False

    return user_passes_test(test)


class PermissionPolicyChecker:
    """
    Provides a view decorator that enforces the given permission policy,
    returning the wagtailadmin 'permission denied' response if permission not granted
    """

    def __init__(self, policy):
        self.policy = policy

    def require(self, action):
        def test(user):
            return self.policy.user_has_permission(user, action)

        return user_passes_test(test)

    def require_any(self, *actions):
        def test(user):
            return self.policy.user_has_any_permission(user, actions)

        return user_passes_test(test)


def user_has_any_page_permission(user):
    """
    Check if a user has any permission to add, edit, or otherwise manage any
    page.
    """
    return page_permission_policy.user_has_any_permission(
        user, {"add", "change", "publish", "bulk_delete", "lock", "unlock"}
    )


def reject_request(request):
    if request.headers.get("x-requested-with") == "XMLHttpRequest":
        raise PermissionDenied

    # import redirect_to_login here to avoid circular imports on model files that import
    # wagtail.admin.auth, specifically where custom user models are involved
    from django.contrib.auth.views import redirect_to_login as auth_redirect_to_login

    login_url = getattr(
        settings, "WAGTAILADMIN_LOGIN_URL", reverse("wagtailadmin_login")
    )

    return auth_redirect_to_login(request.get_full_path(), login_url=login_url)


def require_admin_access(view_func):
    def decorated_view(request, *args, **kwargs):
        user = request.user

        if user.is_anonymous:
            return reject_request(request)

        if user.has_perms(["wagtailadmin.access_admin"]):
            try:
                with LogContext(user=user):
                    return get_localized_response(view_func, request, *args, **kwargs)

            except PermissionDenied:
                if request.headers.get("x-requested-with") == "XMLHttpRequest":
                    raise

                return permission_denied(request)

        if not request.headers.get("x-requested-with") == "XMLHttpRequest":
            messages.error(request, _("You do not have permission to access the admin"))

        return reject_request(request)

    return decorated_view



from django.db import migrations
from django.db.models import Count, Q
from wagtail.models import Page as RealPage


def ancestor_of_q(page):
    paths = [page.path[0:pos] for pos in range(0, len(page.path) + 1, page.steplen)[1:]]
    q = Q(path__in=paths)

    return q


def create_default_workflows(apps, schema_editor):
    # This will recreate the existing publish-permission based moderation setup in the new workflow system, by creating new workflows

    # Get models
    ContentType = apps.get_model("contenttypes.ContentType")
    Workflow = apps.get_model("wagtailcore.Workflow")
    GroupApprovalTask = apps.get_model("wagtailcore.GroupApprovalTask")
    GroupPagePermission = apps.get_model("wagtailcore.GroupPagePermission")
    WorkflowPage = apps.get_model("wagtailcore.WorkflowPage")
    WorkflowTask = apps.get_model("wagtailcore.WorkflowTask")
    Page = apps.get_model("wagtailcore.Page")
    Group = apps.get_model("auth.Group")

    # Get this from real page model just in case it has been overridden
    Page.steplen = RealPage.steplen

    # Create content type for GroupApprovalTask model
    group_approval_content_type, __ = ContentType.objects.get_or_create(
        model="groupapprovaltask", app_label="wagtailcore"
    )

    publish_permissions = GroupPagePermission.objects.filter(permission_type="publish")

    for permission in publish_permissions:
        # find groups with publish permission over this page or its ancestors (and therefore this page by descent)
        page = permission.page
        page = Page.objects.get(pk=page.pk)
        ancestors = Page.objects.filter(ancestor_of_q(page))
        ancestor_permissions = publish_permissions.filter(page__in=ancestors)
        groups = Group.objects.filter(
            Q(page_permissions__in=ancestor_permissions)
            | Q(page_permissions__pk=permission.pk)
        ).distinct()

        # get a GroupApprovalTask with groups matching these publish permission groups (and no others)
        task = (
            GroupApprovalTask.objects.filter(groups__id__in=groups.all())
            .annotate(count=Count("groups"))
            .filter(count=groups.count())
            .filter(active=True)
            .first()
        )
        if not task:
            # if no such task exists, create it
            group_names = " ".join([group.name for group in groups])
            task = GroupApprovalTask.objects.create(
                name=group_names + " approval",
                content_type=group_approval_content_type,
                active=True,
            )
            task.groups.set(groups)

        # get a Workflow containing only this task if it exists, otherwise create it
        workflow = (
            Workflow.objects.annotate(task_number=Count("workflow_tasks"))
            .filter(task_number=1)
            .filter(workflow_tasks__task=task)
            .filter(active=True)
            .first()
        )
        if not workflow:
            workflow = Workflow.objects.create(name=task.name, active=True)

            WorkflowTask.objects.create(
                workflow=workflow,
                task=task,
                sort_order=0,
            )

        # if the workflow is not linked by a WorkflowPage to the permission's linked page, link it by creating a new WorkflowPage now
        if not WorkflowPage.objects.filter(workflow=workflow, page=page).exists():
            WorkflowPage.objects.create(workflow=workflow, page=page)


class Migration(migrations.Migration):

    dependencies = [
        ("wagtailcore", "0047_add_workflow_models"),
    ]

    operations = [
        migrations.RunPython(create_default_workflows, migrations.RunPython.noop),
    ]



from unittest import mock
from urllib.error import HTTPError, URLError

import requests
from azure.mgmt.cdn import CdnManagementClient
from azure.mgmt.frontdoor import FrontDoorManagementClient
from django.core.exceptions import ImproperlyConfigured
from django.test import SimpleTestCase, TestCase
from django.test.utils import override_settings

from wagtail.contrib.frontend_cache.backends import (
    AzureCdnBackend,
    AzureFrontDoorBackend,
    BaseBackend,
    CloudflareBackend,
    CloudfrontBackend,
    HTTPBackend,
)
from wagtail.contrib.frontend_cache.utils import get_backends
from wagtail.models import Page
from wagtail.test.testapp.models import EventIndex, EventPage

from .utils import (
    PurgeBatch,
    purge_page_from_cache,
    purge_pages_from_cache,
    purge_url_from_cache,
    purge_urls_from_cache,
)

EVENTPAGE_URLS = {
    "http://localhost/events/final-event/",
    "http://localhost/events/christmas/",
    "http://localhost/events/saint-patrick/",
    "http://localhost/events/tentative-unpublished-event/",
    "http://localhost/events/someone-elses-event/",
    "http://localhost/events/tentative-unpublished-event/",
    "http://localhost/secret-plans/steal-underpants/",
}


class TestBackendConfiguration(SimpleTestCase):
    def test_default(self):
        backends = get_backends()

        self.assertEqual(len(backends), 0)

    def test_varnish(self):
        backends = get_backends(
            backend_settings={
                "varnish": {
                    "BACKEND": "wagtail.contrib.frontend_cache.backends.HTTPBackend",
                    "LOCATION": "http://localhost:8000",
                },
            }
        )

        self.assertEqual(set(backends.keys()), {"varnish"})
        self.assertIsInstance(backends["varnish"], HTTPBackend)

        self.assertEqual(backends["varnish"].cache_scheme, "http")
        self.assertEqual(backends["varnish"].cache_netloc, "localhost:8000")

    def test_cloudflare(self):
        backends = get_backends(
            backend_settings={
                "cloudflare": {
                    "BACKEND": "wagtail.contrib.frontend_cache.backends.CloudflareBackend",
                    "EMAIL": "test@test.com",
                    "API_KEY": "this is the api key",
                    "ZONEID": "this is a zone id",
                    "BEARER_TOKEN": "this is a bearer token",
                },
            }
        )

        self.assertEqual(set(backends.keys()), {"cloudflare"})
        self.assertIsInstance(backends["cloudflare"], CloudflareBackend)

        self.assertEqual(backends["cloudflare"].cloudflare_email, "test@test.com")
        self.assertEqual(
            backends["cloudflare"].cloudflare_api_key, "this is the api key"
        )
        self.assertEqual(
            backends["cloudflare"].cloudflare_token, "this is a bearer token"
        )

    def test_cloudfront(self):
        backends = get_backends(
            backend_settings={
                "cloudfront": {
                    "BACKEND": "wagtail.contrib.frontend_cache.backends.CloudfrontBackend",
                    "DISTRIBUTION_ID": "frontend",
                    "AWS_ACCESS_KEY_ID": "my-access-key-id",
                    "AWS_SECRET_ACCESS_KEY": "my-secret-access-key",
                },
            }
        )

        self.assertEqual(set(backends.keys()), {"cloudfront"})
        self.assertIsInstance(backends["cloudfront"], CloudfrontBackend)

        self.assertEqual(backends["cloudfront"].cloudfront_distribution_id, "frontend")

        credentials = backends["cloudfront"].client._request_signer._credentials

        self.assertEqual(credentials.method, "explicit")
        self.assertEqual(credentials.access_key, "my-access-key-id")
        self.assertEqual(credentials.secret_key, "my-secret-access-key")

    def test_azure_cdn(self):
        backends = get_backends(
            backend_settings={
                "azure_cdn": {
                    "BACKEND": "wagtail.contrib.frontend_cache.backends.AzureCdnBackend",
                    "RESOURCE_GROUP_NAME": "test-resource-group",
                    "CDN_PROFILE_NAME": "wagtail-io-profile",
                    "CDN_ENDPOINT_NAME": "wagtail-io-endpoint",
                },
            }
        )

        self.assertEqual(set(backends.keys()), {"azure_cdn"})
        self.assertIsInstance(backends["azure_cdn"], AzureCdnBackend)
        self.assertEqual(
            backends["azure_cdn"]._resource_group_name, "test-resource-group"
        )
        self.assertEqual(backends["azure_cdn"]._cdn_profile_name, "wagtail-io-profile")
        self.assertEqual(
            backends["azure_cdn"]._cdn_endpoint_name, "wagtail-io-endpoint"
        )

    def test_azure_front_door(self):
        backends = get_backends(
            backend_settings={
                "azure_front_door": {
                    "BACKEND": "wagtail.contrib.frontend_cache.backends.AzureFrontDoorBackend",
                    "RESOURCE_GROUP_NAME": "test-resource-group",
                    "FRONT_DOOR_NAME": "wagtail-io-front-door",
                },
            }
        )

        self.assertEqual(set(backends.keys()), {"azure_front_door"})
        self.assertIsInstance(backends["azure_front_door"], AzureFrontDoorBackend)
        self.assertEqual(
            backends["azure_front_door"]._resource_group_name, "test-resource-group"
        )
        self.assertEqual(
            backends["azure_front_door"]._front_door_name, "wagtail-io-front-door"
        )

    def test_azure_cdn_get_client(self):
        mock_credentials = mock.MagicMock()
        backends = get_backends(
            backend_settings={
                "azure_cdn": {
                    "BACKEND": "wagtail.contrib.frontend_cache.backends.AzureCdnBackend",
                    "RESOURCE_GROUP_NAME": "test-resource-group",
                    "CDN_PROFILE_NAME": "wagtail-io-profile",
                    "CDN_ENDPOINT_NAME": "wagtail-io-endpoint",
                    "SUBSCRIPTION_ID": "fake-subscription-id",
                    "CREDENTIALS": mock_credentials,
                },
            }
        )
        self.assertEqual(set(backends.keys()), {"azure_cdn"})
        client = backends["azure_cdn"]._get_client()
        self.assertIsInstance(client, CdnManagementClient)
        self.assertEqual(client._config.subscription_id, "fake-subscription-id")
        self.assertIs(client._config.credential, mock_credentials)

    def test_azure_front_door_get_client(self):
        mock_credentials = mock.MagicMock()
        backends = get_backends(
            backend_settings={
                "azure_front_door": {
                    "BACKEND": "wagtail.contrib.frontend_cache.backends.AzureFrontDoorBackend",
                    "RESOURCE_GROUP_NAME": "test-resource-group",
                    "FRONT_DOOR_NAME": "wagtail-io-fake-front-door-name",
                    "SUBSCRIPTION_ID": "fake-subscription-id",
                    "CREDENTIALS": mock_credentials,
                },
            }
        )
        client = backends["azure_front_door"]._get_client()
        self.assertEqual(set(backends.keys()), {"azure_front_door"})
        self.assertIsInstance(client, FrontDoorManagementClient)
        self.assertEqual(client._config.subscription_id, "fake-subscription-id")
        self.assertIs(client._config.credential, mock_credentials)

    @mock.patch(
        "wagtail.contrib.frontend_cache.backends.azure.AzureCdnBackend._make_purge_call"
    )
    def test_azure_cdn_purge(self, make_purge_call_mock):
        backends = get_backends(
            backend_settings={
                "azure_cdn": {
                    "BACKEND": "wagtail.contrib.frontend_cache.backends.AzureCdnBackend",
                    "RESOURCE_GROUP_NAME": "test-resource-group",
                    "CDN_PROFILE_NAME": "wagtail-io-profile",
                    "CDN_ENDPOINT_NAME": "wagtail-io-endpoint",
                    "CREDENTIALS": "Fake credentials",
                },
            }
        )

        self.assertEqual(set(backends.keys()), {"azure_cdn"})
        self.assertIsInstance(backends["azure_cdn"], AzureCdnBackend)

        # purge()
        backends["azure_cdn"].purge(
            "http://www.wagtail.org/home/events/christmas/?test=1"
        )
        make_purge_call_mock.assert_called_once()
        call_args = tuple(make_purge_call_mock.call_args)[0]
        self.assertEqual(len(call_args), 2)
        self.assertIsInstance(call_args[0], CdnManagementClient)
        self.assertEqual(call_args[1], ["/home/events/christmas/?test=1"])
        make_purge_call_mock.reset_mock()

        # purge_batch()
        backends["azure_cdn"].purge_batch(
            [
                "http://www.wagtail.org/home/events/christmas/?test=1",
                "http://torchbox.com/blog/",
            ]
        )
        make_purge_call_mock.assert_called_once()
        call_args = tuple(make_purge_call_mock.call_args)[0]
        self.assertIsInstance(call_args[0], CdnManagementClient)
        self.assertEqual(call_args[1], ["/home/events/christmas/?test=1", "/blog/"])

    @mock.patch(
        "wagtail.contrib.frontend_cache.backends.azure.AzureFrontDoorBackend._make_purge_call"
    )
    def test_azure_front_door_purge(self, make_purge_call_mock):
        backends = get_backends(
            backend_settings={
                "azure_front_door": {
                    "BACKEND": "wagtail.contrib.frontend_cache.backends.AzureFrontDoorBackend",
                    "RESOURCE_GROUP_NAME": "test-resource-group",
                    "FRONT_DOOR_NAME": "wagtail-io-front-door",
                    "CREDENTIALS": "Fake credentials",
                },
            }
        )

        self.assertEqual(set(backends.keys()), {"azure_front_door"})
        self.assertIsInstance(backends["azure_front_door"], AzureFrontDoorBackend)

        # purge()
        backends["azure_front_door"].purge(
            "http://www.wagtail.org/home/events/christmas/?test=1"
        )
        make_purge_call_mock.assert_called_once()
        call_args = tuple(make_purge_call_mock.call_args)[0]
        self.assertIsInstance(call_args[0], FrontDoorManagementClient)
        self.assertEqual(call_args[1], ["/home/events/christmas/?test=1"])

        make_purge_call_mock.reset_mock()

        # purge_batch()
        backends["azure_front_door"].purge_batch(
            [
                "http://www.wagtail.org/home/events/christmas/?test=1",
                "http://torchbox.com/blog/",
            ]
        )
        make_purge_call_mock.assert_called_once()
        call_args = tuple(make_purge_call_mock.call_args)[0]
        self.assertIsInstance(call_args[0], FrontDoorManagementClient)
        self.assertEqual(call_args[1], ["/home/events/christmas/?test=1", "/blog/"])

    def test_http(self):
        """Test that `HTTPBackend.purge` works when urlopen succeeds"""
        self._test_http_with_side_effect(urlopen_side_effect=None)

    def test_http_httperror(self):
        """Test that `HTTPBackend.purge` can handle `HTTPError`"""
        http_error = HTTPError(
            url="http://localhost:8000/home/events/christmas/",
            code=500,
            msg="Internal Server Error",
            hdrs={},
            fp=None,
        )
        with self.assertLogs(level="ERROR") as log_output:
            self._test_http_with_side_effect(urlopen_side_effect=http_error)

        self.assertIn(
            "Couldn't purge 'http://www.wagtail.org/home/events/christmas/' from HTTP cache. HTTPError: 500 Internal Server Error",
            log_output.output[0],
        )

    def test_http_urlerror(self):
        """Test that `HTTPBackend.purge` can handle `URLError`"""
        url_error = URLError(reason="just for tests")
        with self.assertLogs(level="ERROR") as log_output:
            self._test_http_with_side_effect(urlopen_side_effect=url_error)
        self.assertIn(
            "Couldn't purge 'http://www.wagtail.org/home/events/christmas/' from HTTP cache. URLError: just for tests",
            log_output.output[0],
        )

    @mock.patch("wagtail.contrib.frontend_cache.backends.http.urlopen")
    def _test_http_with_side_effect(self, urlopen_mock, urlopen_side_effect):
        # given a backends configuration with one HTTP backend
        backends = get_backends(
            backend_settings={
                "varnish": {
                    "BACKEND": "wagtail.contrib.frontend_cache.backends.HTTPBackend",
                    "LOCATION": "http://localhost:8000",
                },
            }
        )
        self.assertEqual(set(backends.keys()), {"varnish"})
        self.assertIsInstance(backends["varnish"], HTTPBackend)
        # and mocked urlopen that may or may not raise network-related exception
        urlopen_mock.side_effect = urlopen_side_effect

        # when making a purge request
        backends.get("varnish").purge("http://www.wagtail.org/home/events/christmas/")

        # then no exception is raised
        # and mocked urlopen is called with a proper purge request
        self.assertEqual(urlopen_mock.call_count, 1)
        (purge_request,), _call_kwargs = urlopen_mock.call_args
        self.assertEqual(
            purge_request.full_url, "http://localhost:8000/home/events/christmas/"
        )

    def test_cloudfront_validate_distribution_id(self):
        with self.assertRaises(ImproperlyConfigured):
            get_backends(
                backend_settings={
                    "cloudfront": {
                        "BACKEND": "wagtail.contrib.frontend_cache.backends.CloudfrontBackend",
                    },
                }
            )

    @mock.patch(
        "wagtail.contrib.frontend_cache.backends.cloudfront.CloudfrontBackend._create_invalidation"
    )
    def test_cloudfront_purge_with_query_string(self, mock_create_invalidation):
        """Test that CloudFront invalidation includes query strings when purging URLs."""
        backends = get_backends(
            backend_settings={
                "cloudfront": {
                    "BACKEND": "wagtail.contrib.frontend_cache.backends.CloudfrontBackend",
                    "DISTRIBUTION_ID": "frontend",
                    "AWS_ACCESS_KEY_ID": "test-access-key",
                    "AWS_SECRET_ACCESS_KEY": "test-secret-key",
                },
            }
        )

        backends["cloudfront"].purge(
            "http://www.example.com/path/to/page?query=value&another=param"
        )
        mock_create_invalidation.assert_called_once_with(
            "frontend", ["/path/to/page?query=value&another=param"]
        )

    @mock.patch(
        "wagtail.contrib.frontend_cache.backends.cloudfront.CloudfrontBackend._create_invalidation"
    )
    def test_cloudfront_purge_root_with_query_string(self, mock_create_invalidation):
        """Test that CloudFront invalidation handles root URLs with query strings."""
        backends = get_backends(
            backend_settings={
                "cloudfront": {
                    "BACKEND": "wagtail.contrib.frontend_cache.backends.CloudfrontBackend",
                    "DISTRIBUTION_ID": "frontend",
                    "AWS_ACCESS_KEY_ID": "test-access-key",
                    "AWS_SECRET_ACCESS_KEY": "test-secret-key",
                },
            }
        )

        backends["cloudfront"].purge("http://www.example.com?query=value")
        mock_create_invalidation.assert_called_once_with("frontend", ["/?query=value"])

    @mock.patch(
        "wagtail.contrib.frontend_cache.backends.cloudfront.CloudfrontBackend._create_invalidation"
    )
    def test_cloudfront_purge_batch_with_query_strings(self, mock_create_invalidation):
        """Test that CloudFront invalidation batch includes query strings."""
        backends = get_backends(
            backend_settings={
                "cloudfront": {
                    "BACKEND": "wagtail.contrib.frontend_cache.backends.CloudfrontBackend",
                    "DISTRIBUTION_ID": "frontend",
                    "AWS_ACCESS_KEY_ID": "test-access-key",
                    "AWS_SECRET_ACCESS_KEY": "test-secret-key",
                },
            }
        )

        backends["cloudfront"].purge_batch(
            [
                "http://www.example.com/path/to/page?query=value",
                "http://www.example.com/another/path",
                "http://www.example.com?root=query",
            ]
        )
        mock_create_invalidation.assert_called_once()
        self.assertEqual(mock_create_invalidation.call_args.args[0], "frontend")
        self.assertEqual(
            sorted(mock_create_invalidation.call_args.args[1]),
            ["/?root=query", "/another/path", "/path/to/page?query=value"],
        )

    def test_multiple(self):
        backends = get_backends(
            backend_settings={
                "varnish": {
                    "BACKEND": "wagtail.contrib.frontend_cache.backends.HTTPBackend",
                    "LOCATION": "http://localhost:8000/",
                },
                "cloudflare": {
                    "BACKEND": "wagtail.contrib.frontend_cache.backends.CloudflareBackend",
                    "EMAIL": "test@test.com",
                    "API_KEY": "this is the api key",
                    "ZONEID": "this is a zone id",
                },
            }
        )

        self.assertEqual(set(backends.keys()), {"varnish", "cloudflare"})

    def test_filter(self):
        backends = get_backends(
            backend_settings={
                "varnish": {
                    "BACKEND": "wagtail.contrib.frontend_cache.backends.HTTPBackend",
                    "LOCATION": "http://localhost:8000/",
                },
                "cloudflare": {
                    "BACKEND": "wagtail.contrib.frontend_cache.backends.CloudflareBackend",
                    "EMAIL": "test@test.com",
                    "API_KEY": "this is the api key",
                    "ZONEID": "this is a zone id",
                },
            },
            backends=["cloudflare"],
        )

        self.assertEqual(set(backends.keys()), {"cloudflare"})

    @override_settings(WAGTAILFRONTENDCACHE_LOCATION="http://localhost:8000")
    def test_backwards_compatibility(self):
        backends = get_backends()

        self.assertEqual(set(backends.keys()), {"default"})
        self.assertIsInstance(backends["default"], HTTPBackend)
        self.assertEqual(backends["default"].cache_scheme, "http")
        self.assertEqual(backends["default"].cache_netloc, "localhost:8000")


PURGED_URLS = set()


class MockBackend(BaseBackend):
    def purge(self, url):
        PURGED_URLS.add(url)


class MockCloudflareBackend(CloudflareBackend):
    def _purge_urls(self, urls):
        if len(urls) > self.CHUNK_SIZE:
            raise Exception("Cloudflare backend is not chunking requests as expected")

        PURGED_URLS.update(urls)


@override_settings(
    WAGTAILFRONTENDCACHE={
        "varnish": {
            "BACKEND": "wagtail.contrib.frontend_cache.tests.MockBackend",
        },
    },
    CACHES={
        "default": {
            "BACKEND": "django.core.cache.backends.dummy.DummyCache",
        }
    },
)
class TestCachePurgingFunctions(TestCase):
    fixtures = ["test.json"]

    def setUp(self):
        PURGED_URLS.clear()

    def test_purge_url_from_cache(self):
        with self.captureOnCommitCallbacks(execute=True):
            purge_url_from_cache("http://localhost/foo")
        self.assertEqual(PURGED_URLS, {"http://localhost/foo"})

    def test_purge_urls_from_cache(self):
        with self.captureOnCommitCallbacks(execute=True):
            purge_urls_from_cache(["http://localhost/foo", "http://localhost/bar"])
        self.assertEqual(PURGED_URLS, {"http://localhost/foo", "http://localhost/bar"})

    def test_purge_page_from_cache(self):
        page = EventIndex.objects.get(url_path="/home/events/")
        with self.captureOnCommitCallbacks(execute=True):
            with self.assertNumQueries(1):
                # Because no cache object is provided, a query is needed to
                # fetch site root paths in order to derive page urls
                purge_page_from_cache(page)
        self.assertEqual(
            PURGED_URLS,
            {"http://localhost/events/", "http://localhost/events/past/"},
        )

    def test_purge_page_from_cache_with_shared_cache_object(self):
        page = EventIndex.objects.get(url_path="/home/events/")

        # Ensure site root paths are already cached, which should result in
        # zero additional queries being incurred by this test
        page._get_relevant_site_root_paths()

        with self.captureOnCommitCallbacks(execute=True):
            # Because site root paths are already available via the cache_object,
            # no further queries should be needed to derive page urls
            with self.assertNumQueries(0):
                purge_page_from_cache(page, cache_object=page)

        self.assertEqual(
            PURGED_URLS,
            {"http://localhost/events/", "http://localhost/events/past/"},
        )

    def test_purge_pages_from_cache(self):
        pages = list(Page.objects.all().type(EventPage))
        with self.captureOnCommitCallbacks(execute=True):
            with self.assertNumQueries(1):
                # Because no cache object is provided, a query is needed to
                # fetch site root paths in order to derive page urls
                purge_pages_from_cache(pages)
        self.assertEqual(PURGED_URLS, EVENTPAGE_URLS)

    def test_purge_pages_from_cache_with_shared_cache_object(self):
        pages = list(Page.objects.all().type(EventPage))

        # Use the first page as the cache object for the operation
        cache_object = pages[0]

        # Ensure site root paths are already cached, which should result in
        # zero additional queries being incurred by this test
        cache_object._get_relevant_site_root_paths()

        with self.captureOnCommitCallbacks(execute=True):
            # Because site root paths are already available via the cache_object,
            # no further queries should be needed to derive page urls
            with self.assertNumQueries(0):
                purge_pages_from_cache(pages, cache_object=cache_object)

        self.assertEqual(PURGED_URLS, EVENTPAGE_URLS)

    def test_purge_batch(self):
        page = EventIndex.objects.get(url_path="/home/events/")
        batch = PurgeBatch()

        # Because no cache object is provided, a query is needed to
        # fetch site root paths in order to derive page urls
        with self.assertNumQueries(1):
            batch.add_page(page)

        batch.add_url("http://localhost/foo")

        with self.captureOnCommitCallbacks(execute=True):
            batch.purge()

        self.assertEqual(
            PURGED_URLS,
            {
                "http://localhost/events/",
                "http://localhost/events/past/",
                "http://localhost/foo",
            },
        )

    def test_purge_batch_with_multiple_pages(self):
        pages = list(Page.objects.all().type(EventPage))
        batch = PurgeBatch()

        # Because the batch has no cache object, a query is needed to
        # fetch site root paths in order to derive page urls
        with self.assertNumQueries(1):
            batch.add_pages(pages)

        with self.captureOnCommitCallbacks(execute=True):
            batch.purge()

        self.assertEqual(PURGED_URLS, EVENTPAGE_URLS)

    def test_multiple_purge_batches_with_shared_cache_object(self):
        pages = list(Page.objects.all().type(EventPage))

        # Use the first page as the cache object for the batch
        cache_object = pages[0]

        # Ensure site root paths are already cached, which should result in
        # zero additional queries being incurred by this test
        cache_object._get_relevant_site_root_paths()

        batch = PurgeBatch(cache_object=cache_object)

        with self.assertNumQueries(0):
            # Because site root paths are already available via the cache_object,
            # no queries should be needed to derive page urls
            batch.add_pages(pages)

        with self.captureOnCommitCallbacks(execute=True):
            batch.purge()

        self.assertEqual(PURGED_URLS, EVENTPAGE_URLS)
        PURGED_URLS.clear()

    @override_settings(
        WAGTAILFRONTENDCACHE={
            "varnish": {
                "BACKEND": "wagtail.contrib.frontend_cache.tests.MockBackend",
                "HOSTNAMES": ["example.com"],
            },
        }
    )
    def test_invalidate_specific_location(self):
        with self.assertLogs(level="INFO") as log_output:
            with self.captureOnCommitCallbacks(execute=True):
                purge_url_from_cache("http://localhost/foo")

        self.assertEqual(PURGED_URLS, set())
        self.assertIn(
            "Unable to find purge backend for localhost",
            log_output.output[0],
        )

        with self.captureOnCommitCallbacks(execute=True):
            purge_url_from_cache("http://example.com/foo")
        self.assertEqual(PURGED_URLS, {"http://example.com/foo"})


@override_settings(
    WAGTAILFRONTENDCACHE={
        "cloudflare": {
            "BACKEND": "wagtail.contrib.frontend_cache.tests.MockCloudflareBackend",
            "ZONEID": "zone",
            "BEARER_TOKEN": "token",
        },
    }
)
class TestCloudflareCachePurgingFunctions(TestCase):
    def setUp(self):
        # Reset PURGED_URLS to an empty list
        PURGED_URLS.clear()

    def test_cloudflare_purge_batch_chunked(self):
        with self.captureOnCommitCallbacks(execute=True):
            batch = PurgeBatch()
            urls = [f"https://localhost/foo{i}" for i in range(1, 65)]
            batch.add_urls(urls)
            batch.purge()

        self.assertCountEqual(PURGED_URLS, set(urls))


@override_settings(
    WAGTAILFRONTENDCACHE={
        "varnish": {
            "BACKEND": "wagtail.contrib.frontend_cache.tests.MockBackend",
        },
    }
)
class TestCachePurgingSignals(TestCase):
    fixtures = ["test.json"]

    def setUp(self):
        # Reset PURGED_URLS to an empty list
        PURGED_URLS.clear()

    def test_purge_on_publish(self):
        with self.captureOnCommitCallbacks(execute=True):
            page = EventIndex.objects.get(url_path="/home/events/")
            page.save_revision().publish()
        self.assertEqual(
            PURGED_URLS, {"http://localhost/events/", "http://localhost/events/past/"}
        )

    def test_purge_on_unpublish(self):
        with self.captureOnCommitCallbacks(execute=True):
            page = EventIndex.objects.get(url_path="/home/events/")
            page.unpublish()
        self.assertEqual(
            PURGED_URLS, {"http://localhost/events/", "http://localhost/events/past/"}
        )

    def test_purge_with_unroutable_page(self):
        with self.captureOnCommitCallbacks(execute=True):
            root = Page.objects.get(url_path="/")
            page = EventIndex(title="new top-level page")
            root.add_child(instance=page)
            page.save_revision().publish()
        self.assertEqual(PURGED_URLS, set())

    @override_settings(
        ROOT_URLCONF="wagtail.test.urls_multilang",
        LANGUAGE_CODE="en",
        WAGTAILFRONTENDCACHE_LANGUAGES=["en", "fr", "pt-br"],
    )
    def test_purge_on_publish_in_multilang_env(self):
        with self.captureOnCommitCallbacks(execute=True):
            page = EventIndex.objects.get(url_path="/home/events/")
            page.save_revision().publish()

        self.assertEqual(
            PURGED_URLS,
            {
                "http://localhost/en/events/",
                "http://localhost/en/events/past/",
                "http://localhost/fr/events/",
                "http://localhost/fr/events/past/",
                "http://localhost/pt-br/events/",
                "http://localhost/pt-br/events/past/",
            },
        )

    @override_settings(
        ROOT_URLCONF="wagtail.test.urls_multilang",
        LANGUAGE_CODE="en",
        WAGTAIL_I18N_ENABLED=True,
        WAGTAIL_CONTENT_LANGUAGES=[("en", "English"), ("fr", "French")],
    )
    def test_purge_on_publish_with_i18n_enabled(self):
        with self.captureOnCommitCallbacks(execute=True):
            page = EventIndex.objects.get(url_path="/home/events/")
            page.save_revision().publish()

        self.assertEqual(
            PURGED_URLS,
            {
                "http://localhost/en/events/",
                "http://localhost/en/events/past/",
                "http://localhost/fr/events/",
                "http://localhost/fr/events/past/",
            },
        )

    @override_settings(
        ROOT_URLCONF="wagtail.test.urls_multilang",
        LANGUAGE_CODE="en",
        WAGTAIL_CONTENT_LANGUAGES=[("en", "English"), ("fr", "French")],
    )
    def test_purge_on_publish_without_i18n_enabled(self):
        with self.captureOnCommitCallbacks(execute=True):
            # It should ignore WAGTAIL_CONTENT_LANGUAGES as WAGTAIL_I18N_ENABLED isn't set
            page = EventIndex.objects.get(url_path="/home/events/")
            page.save_revision().publish()
        self.assertEqual(
            PURGED_URLS,
            {"http://localhost/en/events/", "http://localhost/en/events/past/"},
        )


class TestPurgeBatchClass(TestCase):
    # Tests the .add_*() methods on PurgeBatch. The .purge() method is tested
    # by TestCachePurgingFunctions.test_purge_batch above

    fixtures = ["test.json"]

    def test_add_url(self):
        batch = PurgeBatch()
        batch.add_url("http://localhost/foo")

        self.assertEqual(batch.urls, {"http://localhost/foo"})

    def test_add_urls(self):
        batch = PurgeBatch()
        batch.add_urls(["http://localhost/foo", "http://localhost/bar"])

        self.assertEqual(batch.urls, {"http://localhost/foo", "http://localhost/bar"})

    def test_add_page(self):
        page = EventIndex.objects.get(url_path="/home/events/")

        batch = PurgeBatch()
        batch.add_page(page)

        self.assertEqual(
            batch.urls, {"http://localhost/events/", "http://localhost/events/past/"}
        )

    def test_add_pages(self):
        batch = PurgeBatch()
        batch.add_pages(EventIndex.objects.all())

        self.assertEqual(
            batch.urls, {"http://localhost/events/", "http://localhost/events/past/"}
        )

    def test_multiple_calls(self):
        page = EventIndex.objects.get(url_path="/home/events/")

        batch = PurgeBatch()
        batch.add_page(page)
        batch.add_url("http://localhost/foo")
        batch.purge()

        self.assertEqual(
            batch.urls,
            {
                "http://localhost/events/",
                "http://localhost/events/past/",
                "http://localhost/foo",
            },
        )

    @mock.patch("wagtail.contrib.frontend_cache.backends.cloudflare.requests.delete")
    def test_http_error_on_cloudflare_purge_batch(self, requests_delete_mock):
        backend_settings = {
            "cloudflare": {
                "BACKEND": "wagtail.contrib.frontend_cache.backends.CloudflareBackend",
                "EMAIL": "test@test.com",
                "API_KEY": "this is the api key",
                "ZONEID": "this is a zone id",
            },
        }

        class MockResponse:
            def __init__(self, status_code=200):
                self.status_code = status_code

        http_error = requests.exceptions.HTTPError(
            response=MockResponse(status_code=500)
        )
        requests_delete_mock.side_effect = http_error

        batch = PurgeBatch()
        batch.add_url("http://localhost/events/")

        with self.assertLogs(level="ERROR") as log_output:
            with self.captureOnCommitCallbacks(execute=True):
                batch.purge(backend_settings=backend_settings)

        self.assertIn(
            "Couldn't purge 'http://localhost/events/' from Cloudflare. HTTPError: 500",
            log_output.output[0],
        )



from django.test import TestCase

from wagtail.images.models import Image
from wagtail.images.tests.utils import (
    get_test_image_file,
    get_test_image_file_svg,
    get_test_image_filename,
)
from wagtail.test.utils import WagtailTestUtils


class TestJinja2SVGSupport(WagtailTestUtils, TestCase):
    """Test SVG support in Jinja2 templates with preserve-svg filter."""

    def setUp(self):
        # Create a real test engine
        from django.template.loader import engines

        self.engine = engines["jinja2"]

        # Create a raster image
        self.raster_image = Image.objects.create(
            title="Test raster image",
            file=get_test_image_file(),
        )

        # Create an SVG image
        self.svg_image = Image.objects.create(
            title="Test SVG image",
            file=get_test_image_file_svg(),
        )

    def render(self, string, context=None):
        if context is None:
            context = {}

        template = self.engine.from_string(string)
        return template.render(context)

    def test_image_with_raster_image(self):
        """Test that raster images work normally without preserve-svg."""
        html = self.render(
            '{{ image(img, "width-200|format-webp") }}', {"img": self.raster_image}
        )
        filename = get_test_image_filename(self.raster_image, "width-200.format-webp")

        self.assertHTMLEqual(
            html,
            f'<img src="{filename}" width="200" height="150" alt="Test raster image">',
        )

    def test_image_with_svg_without_preserve(self):
        """Test that without preserve-svg, SVGs get all operations (which would fail in production)."""
        with self.assertRaises(AttributeError):
            self.render(
                '{{ image(img, "width-200|format-webp") }}', {"img": self.svg_image}
            )

    def test_image_with_svg_with_preserve(self):
        """Test that with preserve-svg filter, SVGs only get safe operations."""
        html = self.render(
            '{{ image(img, "width-45|format-webp|preserve-svg") }}',
            {"img": self.svg_image},
        )
        filename = get_test_image_filename(self.svg_image, "width-45")

        self.assertHTMLEqual(
            html,
            f'<img src="{filename}" width="45.0" height="45.0" alt="Test SVG image">',
        )

    def test_raster_image_with_preserve_svg(self):
        """Test that preserve-svg on image tag has no effect on raster images."""
        html = self.render(
            '{{ image(img, "width-100|format-webp|preserve-svg") }}',
            {"img": self.raster_image},
        )
        filename = get_test_image_filename(self.raster_image, "width-100.format-webp")

        self.assertHTMLEqual(
            html,
            f'<img src="{filename}" width="100" height="75" alt="Test raster image">',
        )

    def test_srcset_image_with_svg_preserve(self):
        """Test that preserve-svg works with srcset_image function."""
        html = self.render(
            '{{ srcset_image(img, "width-{35,55}|format-webp|preserve-svg", sizes="100vw") }}',
            {"img": self.svg_image},
        )
        filename35 = get_test_image_filename(self.svg_image, "width-35")
        filename55 = get_test_image_filename(self.svg_image, "width-55")

        self.assertHTMLEqual(
            html,
            f"""
                <img sizes="100vw" src="{filename35}"
                    srcset="{filename35} 35.0w, {filename55} 55.0w" width="35.0" height="35.0"
                    alt="Test SVG image">
            """,
        )

    def test_raster_srcset_image_with_preserve_svg(self):
        """Test that preserve-svg on srcset_image tag has no effect on raster images."""
        html = self.render(
            '{{ srcset_image(img, "width-{100,120}|format-webp|preserve-svg", sizes="100vw") }}',
            {"img": self.raster_image},
        )
        filename100 = get_test_image_filename(
            self.raster_image, "width-100.format-webp"
        )
        filename120 = get_test_image_filename(
            self.raster_image, "width-120.format-webp"
        )

        self.assertHTMLEqual(
            html,
            f"""
                <img sizes="100vw" src="{filename100}"
                    srcset="{filename100} 100w, {filename120} 120w" width="100" height="75"
                    alt="Test raster image">
            """,
        )

    def test_picture_with_svg_preserve(self):
        """Test that preserve-svg works with picture function."""
        html = self.render(
            '{{ picture(img, "format-{avif,webp,jpeg}|width-85|preserve-svg") }}',
            {"img": self.svg_image},
        )
        filename = get_test_image_filename(self.svg_image, "width-85")
        self.assertHTMLEqual(
            html,
            f"""
                <picture>
                    <img src="{filename}" alt="Test SVG image" width="85.0" height="85.0">
                </picture>
            """,
        )

    def test_raster_picture_with_preserve_svg(self):
        """Test that preserve-svg on picture tag has no effect on raster images."""
        html = self.render(
            '{{ picture(img, "width-160|format-{webp,jpeg}|preserve-svg") }}',
            {"img": self.raster_image},
        )
        filename_webp = get_test_image_filename(
            self.raster_image, "width-160.format-webp"
        )
        filename_jpeg = get_test_image_filename(
            self.raster_image, "width-160.format-jpeg"
        )

        self.assertHTMLEqual(
            html,
            f"""
                <picture>
                    <source srcset="{filename_webp}" type="image/webp">
                    <img src="{filename_jpeg}" alt="Test raster image" width="160" height="120">
                </picture>
            """,
        )

    def test_preserve_svg_with_multiple_operations(self):
        """Test preserve-svg with multiple operations, some safe, some unsafe for SVGs."""
        html = self.render(
            '{{ image(img, "width-300|height-200|format-webp|max-100x100|jpegquality-80|preserve-svg") }}',
            {"img": self.svg_image},
        )
        filename = get_test_image_filename(
            self.svg_image, "width-300.height-200.max-100x100"
        )
        self.assertHTMLEqual(
            html,
            f'<img src="{filename}" alt="Test SVG image" width="100.0" height="100.0">',
        )

    def test_preserve_svg_with_custom_attributes(self):
        """Test preserve-svg works with custom HTML attributes."""
        html = self.render(
            '{{ image(img, "width-66|format-webp|preserve-svg", class="my-image", alt="Custom alt") }}',
            {"img": self.svg_image},
        )
        filename = get_test_image_filename(self.svg_image, "width-66")

        self.assertHTMLEqual(
            html,
            f'<img src="{filename}" class="my-image" alt="Custom alt" width="66.0" height="66.0">',
        )

    def test_picture_with_multiple_formats_and_sizes_with_raster_image(self):
        filenames = [
            get_test_image_filename(self.raster_image, "width-160.format-avif"),
            get_test_image_filename(self.raster_image, "width-320.format-avif"),
            get_test_image_filename(self.raster_image, "width-160.format-jpeg"),
            get_test_image_filename(self.raster_image, "width-320.format-jpeg"),
        ]

        rendered = self.render(
            '{{ picture(myimage, "width-{160,320}|format-{avif,jpeg}|preserve-svg", sizes="100vw") }}',
            {"myimage": self.raster_image},
        )
        expected = f"""
            <picture>
            <source srcset="{filenames[0]} 160w, {filenames[1]} 320w" sizes="100vw" type="image/avif">
            <img
                sizes="100vw"
                src="{filenames[2]}"
                srcset="{filenames[2]} 160w, {filenames[3]} 320w"
                alt="Test raster image"
                width="160"
                height="120"
            >
            </picture>
        """
        self.assertHTMLEqual(rendered, expected)

    def test_picture_with_multiple_formats_and_sizes_with_svg_image(self):
        filenames = [
            get_test_image_filename(self.svg_image, "width-45"),
            get_test_image_filename(self.svg_image, "width-90"),
        ]

        rendered = self.render(
            '{{ picture(myimage, "width-{45,90}|format-{avif,jpeg}|preserve-svg", sizes="100vw") }}',
            {"myimage": self.svg_image},
        )
        expected = f"""
            <picture>
            <img
                sizes="100vw"
                src="{filenames[0]}"
                srcset="{filenames[0]} 45.0w, {filenames[1]} 90.0w"
                alt="Test SVG image"
                width="45.0"
                height="45.0"
            >
            </picture>
        """
        self.assertHTMLEqual(rendered, expected)



from django.apps import apps
from django.core import checks
from django.db import models
from django.test import TestCase

from wagtail.models import DraftStateMixin, LockableMixin, RevisionMixin, WorkflowMixin


class TestWorkflowMixin(TestCase):
    def tearDown(self):
        # Unregister the models from the overall model registry
        # so that it doesn't break tests elsewhere.
        # We can probably replace this with Django's @isolate_apps decorator.
        for package in ("wagtailcore", "wagtail.tests"):
            try:
                for model in (
                    "workflowwithoutrevisionmodel",
                    "workflowwithoutdraftstatemodel",
                    "workflowincorrectordermodel1",
                    "workflowincorrectordermodel2",
                    "correctworkflowmodel",
                    "correctnotlockableworkflowmodel",
                ):
                    del apps.all_models[package][model]
            except KeyError:
                pass
        apps.clear_cache()

    def test_missing_revision_or_draftstate_mixins(self):
        error = checks.Error(
            "WorkflowMixin requires DraftStateMixin and RevisionMixin (in that order).",
            hint=(
                "Make sure your model's inheritance order is as follows: "
                "WorkflowMixin, DraftStateMixin, RevisionMixin."
            ),
            id="wagtailcore.E006",
        )

        class WorkflowWithoutRevisionModel(WorkflowMixin, models.Model):
            pass

        class WorkflowWithoutDraftStateModel(
            WorkflowMixin, RevisionMixin, models.Model
        ):
            pass

        for model in (WorkflowWithoutRevisionModel, WorkflowWithoutDraftStateModel):
            with self.subTest(model=model):
                error.obj = model
                self.assertEqual(model.check(), [error])

    def test_incorrect_mixins_order(self):
        error = checks.Error(
            "WorkflowMixin requires DraftStateMixin and RevisionMixin (in that order).",
            hint=(
                "Make sure your model's inheritance order is as follows: "
                "WorkflowMixin, DraftStateMixin, RevisionMixin."
            ),
            id="wagtailcore.E006",
        )

        class WorkflowIncorrectOrderModel1(
            DraftStateMixin, WorkflowMixin, RevisionMixin, LockableMixin, models.Model
        ):
            pass

        class WorkflowIncorrectOrderModel2(
            DraftStateMixin, RevisionMixin, WorkflowMixin, models.Model
        ):
            pass

        for model in (WorkflowIncorrectOrderModel1, WorkflowIncorrectOrderModel2):
            with self.subTest(model=model):
                error.obj = model
                self.assertEqual(model.check(), [error])

    def test_correct_mixins_order(self):
        class CorrectWorkflowModel(
            WorkflowMixin, DraftStateMixin, LockableMixin, RevisionMixin, models.Model
        ):
            pass

        class CorrectNotLockableWorkflowModel(
            WorkflowMixin, DraftStateMixin, RevisionMixin, models.Model
        ):
            pass

        for model in (CorrectWorkflowModel, CorrectNotLockableWorkflowModel):
            with self.subTest(model=model):
                self.assertEqual(model.check(), [])



from django.conf import settings
from django.core.exceptions import ImproperlyConfigured


def get_document_model_string():
    """
    Get the dotted ``app.Model`` name for the document model as a string.
    Useful for developers making Wagtail plugins that need to refer to the
    document model, such as in foreign keys, but the model itself is not required.
    """
    return getattr(settings, "WAGTAILDOCS_DOCUMENT_MODEL", "wagtaildocs.Document")


def get_document_model():
    """
    Get the document model from the ``WAGTAILDOCS_DOCUMENT_MODEL`` setting.
    Defaults to the standard ``wagtail.documents.models.Document`` model
    if no custom model is defined.
    """
    from django.apps import apps

    model_string = get_document_model_string()
    try:
        return apps.get_model(model_string, require_ready=False)
    except ValueError:
        raise ImproperlyConfigured(
            "WAGTAILDOCS_DOCUMENT_MODEL must be of the form 'app_label.model_name'"
        )
    except LookupError:
        raise ImproperlyConfigured(
            "WAGTAILDOCS_DOCUMENT_MODEL refers to model '%s' that has not been installed"
            % model_string
        )



"""
``wagtail.models`` is split into submodules for maintainability. All definitions intended as
public should be imported here (with 'noqa: F401' comments as required) and outside code should
continue to import them from wagtail.models (e.g. ``from wagtail.models import Site``, not
``from wagtail.models.sites import Site``.)

Submodules should take care to keep the direction of dependencies consistent; where possible they
should implement low-level generic functionality which is then imported by higher-level models such
as Page.
"""

from django.contrib.contenttypes.models import ContentType  # noqa: F401
from modelcluster.fields import ParentalKey  # noqa: F401
from modelcluster.models import ClusterableModel  # noqa: F401
from treebeard.mp_tree import MP_Node  # noqa: F401

from wagtail.query import PageQuerySet  # noqa: F401

from .audit_log import (  # noqa: F401
    BaseLogEntry,
    BaseLogEntryManager,
    LogEntryQuerySet,
    ModelLogEntry,
)
from .content_types import get_default_page_content_type  # noqa: F401
from .copying import _copy, _copy_m2m_relations, _extract_field_data  # noqa: F401
from .draft_state import DraftStateMixin  # noqa: F401
from .i18n import (  # noqa: F401
    BootstrapTranslatableMixin,
    BootstrapTranslatableModel,
    Locale,
    LocaleManager,
    TranslatableMixin,
    bootstrap_translatable_model,
    get_translatable_models,
)
from .locking import LockableMixin  # noqa: F401
from .media import (  # noqa: F401
    BaseCollectionManager,
    Collection,
    CollectionManager,
    CollectionMember,
    CollectionViewRestriction,
    GroupCollectionPermission,
    GroupCollectionPermissionManager,
    UploadedFile,
    get_root_collection_id,
)
from .orderable import Orderable  # noqa: F401
from .pages import (  # noqa: F401
    COMMENTS_RELATION_NAME,
    PAGE_MODEL_CLASSES,
    PAGE_PERMISSION_CODENAMES,
    PAGE_PERMISSION_TYPE_CHOICES,
    PAGE_PERMISSION_TYPES,
    PAGE_TEMPLATE_VAR,
    AbstractPage,
    BasePageManager,
    Comment,
    CommentReply,
    GroupPagePermission,
    GroupPagePermissionManager,
    Page,
    PageBase,
    PageLogEntry,
    PageLogEntryManager,
    PageLogEntryQuerySet,
    PageManager,
    PagePermissionTester,
    PageSubscription,
    PageViewRestriction,
    WorkflowPage,
    get_page_content_types,
    get_page_models,
    get_streamfield_names,
    reassign_root_page_locale_on_delete,
)
from .panels import CommentPanelPlaceholder, PanelPlaceholder  # noqa: F401
from .preview import PreviewableMixin  # noqa: F401
from .reference_index import ReferenceIndex  # noqa: F401
from .revisions import (  # noqa: F401
    PageRevisionsManager,
    Revision,
    RevisionMixin,
    RevisionQuerySet,
    RevisionsManager,
)
from .sites import GroupSitePermission, Site, SiteManager, SiteRootPath  # noqa: F401
from .specific import SpecificMixin  # noqa: F401
from .view_restrictions import BaseViewRestriction  # noqa: F401
from .workflows import (  # noqa: F401
    AbstractGroupApprovalTask,
    AbstractWorkflow,
    BaseTaskStateManager,
    GroupApprovalTask,
    Task,
    TaskManager,
    TaskQuerySet,
    TaskState,
    TaskStateManager,
    TaskStateQuerySet,
    Workflow,
    WorkflowContentType,
    WorkflowManager,
    WorkflowMixin,
    WorkflowState,
    WorkflowStateManager,
    WorkflowStateQuerySet,
    WorkflowTask,
)



from warnings import warn

from django.template.loader import render_to_string
from django.utils import translation
from django.utils.translation import gettext_lazy as _

from wagtail import hooks
from wagtail.admin.ui.components import Component
from wagtail.admin.utils import get_admin_base_url
from wagtail.coreutils import accepts_kwarg
from wagtail.models import Revision
from wagtail.models.pages import Page
from wagtail.users.models import UserProfile
from wagtail.utils.deprecation import RemovedInWagtail80Warning


class BaseItem:
    template = "wagtailadmin/userbar/item_base.html"

    def get_context_data(self, request):
        return {"self": self, "request": request}

    def render(self, request):
        return render_to_string(self.template, self.get_context_data(request))


class AdminItem(BaseItem):
    template = "wagtailadmin/userbar/item_admin.html"


class AccessibilityItem(BaseItem):
    """A userbar item that runs the accessibility checker."""

    def __init__(self, in_editor=False):
        super().__init__()
        self.in_editor = in_editor
        """Whether the accessibility checker is being run in the page editor."""

    #: The template to use for rendering the item.
    template = "wagtailadmin/userbar/item_accessibility.html"

    #: A list of CSS selector(s) to test specific parts of the page.
    #: For more details, see `Axe documentation <https://github.com/dequelabs/axe-core/blob/master/doc/context.md#the-include-property>`__.
    axe_include = ["body"]

    #: A list of CSS selector(s) to exclude specific parts of the page from testing.
    #: For more details, see `Axe documentation <https://github.com/dequelabs/axe-core/blob/master/doc/context.md#exclude-elements-from-test>`__.
    axe_exclude = []

    # Make sure that the userbar is not tested.
    _axe_default_exclude = [{"fromShadowDom": ["wagtail-userbar"]}]

    #: A list of `axe-core tags <https://github.com/dequelabs/axe-core/blob/master/doc/API.md#axe-core-tags>`_
    #: or a list of `axe-core rule IDs <https://github.com/dequelabs/axe-core/blob/master/doc/rule-descriptions.md>`_
    #: (not a mix of both).
    #: Setting this to a falsy value (e.g. ``None``) will omit the ``runOnly`` option and make Axe run with all non-experimental rules enabled.
    axe_run_only = [
        "button-name",
        "empty-heading",
        "empty-table-header",
        "frame-title",
        "heading-order",
        "input-button-name",
        "link-name",
        "p-as-heading",
        "alt-text-quality",
    ]

    #: A dictionary that maps axe-core rule IDs to a dictionary of rule options,
    #: commonly in the format of ``{"enabled": True/False}``. This can be used in
    #: conjunction with :attr:`axe_run_only` to enable or disable specific rules.
    #: For more details, see `Axe documentation <https://github.com/dequelabs/axe-core/blob/master/doc/API.md#options-parameter-examples>`__.
    axe_rules = {}

    #: A list to add custom Axe rules or override their properties,
    #: alongside with ``axe_custom_checks``. Includes Wagtail’s custom rules.
    #: For more details, see `Axe documentation <https://github.com/dequelabs/axe-core/blob/master/doc/API.md#api-name-axeconfigure>`_.
    axe_custom_rules = [
        {
            "id": "alt-text-quality",
            "impact": "serious",
            "selector": "img[alt]",
            "tags": ["best-practice"],
            "any": ["check-image-alt-text"],
            # If omitted, defaults to True and overrides configs in `axe_run_only`.
            "enabled": True,
        },
    ]

    #: A list to add custom Axe checks or override their properties.
    #: Should be used in conjunction with ``axe_custom_rules``.
    #: For more details, see `Axe documentation <https://github.com/dequelabs/axe-core/blob/master/doc/API.md#api-name-axeconfigure>`_.
    axe_custom_checks = [
        {
            "id": "check-image-alt-text",
            "options": {"pattern": "\\.(avif|gif|jpg|jpeg|png|svg|webp)$|_"},
        },
    ]

    #: A dictionary that maps axe-core rule IDs to custom translatable strings
    #: to use as the error messages. If an enabled rule does not exist in this
    #: dictionary, Axe's error message for the rule will be used as fallback.
    axe_messages = {
        "button-name": {
            "error_name": _("Button text is empty"),
            "help_text": _("Use meaningful text for screen reader users"),
        },
        "empty-heading": {
            "error_name": _("Empty heading found"),
            "help_text": _("Use meaningful text for screen reader users"),
        },
        "empty-table-header": {
            "error_name": _("Table header text is empty"),
            "help_text": _("Use meaningful text for screen reader users"),
        },
        "frame-title": {
            "error_name": _("Empty frame title found"),
            "help_text": _("Use a meaningful title for screen reader users"),
        },
        "heading-order": {
            "error_name": _("Incorrect heading hierarchy"),
            "help_text": _("Avoid skipping levels"),
        },
        "input-button-name": {
            "error_name": _("Input button text is empty"),
            "help_text": _("Use meaningful text for screen reader users"),
        },
        "link-name": {
            "error_name": _("Link text is empty"),
            "help_text": _("Use meaningful text for screen reader users"),
        },
        "p-as-heading": {
            "error_name": _("Misusing paragraphs as headings"),
            "help_text": _("Use proper heading tags"),
        },
        "alt-text-quality": {
            "error_name": _("Image alt text has inappropriate pattern"),
            "help_text": _("Use meaningful text"),
        },
    }

    def get_axe_include(self, request):
        """Returns a list of CSS selector(s) to test specific parts of the page."""
        return self.axe_include

    def get_axe_exclude(self, request):
        """Returns a list of CSS selector(s) to exclude specific parts of the page from testing."""
        return self.axe_exclude + self._axe_default_exclude

    def get_axe_run_only(self, request):
        """Returns a list of axe-core tags or a list of axe-core rule IDs (not a mix of both)."""
        return self.axe_run_only

    def get_axe_rules(self, request):
        """Returns a dictionary that maps axe-core rule IDs to a dictionary of rule options."""
        return self.axe_rules

    def get_axe_custom_rules(self, request):
        """List of rule objects per axe.run API."""
        return self.axe_custom_rules

    def get_axe_custom_checks(self, request):
        """List of check objects per axe.run API, without evaluate function."""
        return self.axe_custom_checks

    def get_axe_messages(self, request):
        """Returns a dictionary that maps axe-core rule IDs to custom translatable strings."""
        return self.axe_messages

    def get_axe_context(self, request):
        """
        Returns the `context object <https://github.com/dequelabs/axe-core/blob/develop/doc/context.md>`_
        to be passed as the
        `context parameter <https://github.com/dequelabs/axe-core/blob/develop/doc/API.md#context-parameter>`_
        for ``axe.run``.
        """
        return {
            "include": self.get_axe_include(request),
            "exclude": self.get_axe_exclude(request),
        }

    def get_axe_options(self, request):
        """
        Returns the options object to be passed as the
        `options parameter <https://github.com/dequelabs/axe-core/blob/develop/doc/API.md#options-parameter>`_
        for ``axe.run``.
        """
        options = {
            "runOnly": self.get_axe_run_only(request),
            "rules": self.get_axe_rules(request),
        }
        # If the runOnly option is omitted, Axe will run all rules except those
        # with the "experimental" flag or that are disabled in the rules option.
        # The runOnly has to be omitted (instead of set to an empty list or null)
        # for this to work, so we remove it if it's falsy.
        if not options["runOnly"]:
            options.pop("runOnly")
        return options

    def get_axe_spec(self, request):
        """Returns spec for Axe, including custom rules and custom checks"""
        return {
            "rules": self.get_axe_custom_rules(request),
            "checks": self.get_axe_custom_checks(request),
        }

    def get_axe_configuration(self, request):
        return {
            "context": self.get_axe_context(request),
            "options": self.get_axe_options(request),
            "messages": self.get_axe_messages(request),
            "spec": self.get_axe_spec(request),
        }

    def get_context_data(self, request):
        return {
            **super().get_context_data(request),
            "axe_configuration": self.get_axe_configuration(request),
        }


class AddPageItem(BaseItem):
    template = "wagtailadmin/userbar/item_page_add.html"

    def __init__(self, page):
        self.page = page
        self.parent_page = page.get_parent()

    def render(self, request):
        # Don't render if the page doesn't have an id
        if not self.page.id:
            return ""

        # Don't render if user doesn't have ability to add children here
        permission_checker = self.page.permissions_for_user(request.user)
        if not permission_checker.can_add_subpage():
            return ""

        return super().render(request)


class ExplorePageItem(BaseItem):
    template = "wagtailadmin/userbar/item_page_explore.html"

    def __init__(self, page):
        self.page = page
        self.parent_page = page.get_parent()

    def render(self, request):
        # Don't render if the page doesn't have an id
        if not self.page.id:
            return ""

        # Don't render if user doesn't have ability to edit or publish subpages on the parent page
        permission_checker = self.parent_page.permissions_for_user(request.user)
        if (
            not permission_checker.can_edit()
            and not permission_checker.can_publish_subpage()
        ):
            return ""

        return super().render(request)


class EditPageItem(BaseItem):
    template = "wagtailadmin/userbar/item_page_edit.html"

    def __init__(self, page):
        self.page = page

    def render(self, request):
        # Don't render if the page doesn't have an id
        if not self.page.id:
            return ""

        # Don't render if request is a preview. This is to avoid confusion that
        # might arise when the user clicks edit on a preview.
        try:
            if request.is_preview and request.is_editing:
                return ""
        except AttributeError:
            pass

        # Don't render if the user doesn't have permission to edit this page
        permission_checker = self.page.permissions_for_user(request.user)
        if not permission_checker.can_edit():
            return ""

        return super().render(request)


def apply_userbar_hooks(request, items, page):
    for fn in hooks.get_hooks("construct_wagtail_userbar"):
        if accepts_kwarg(fn, "page"):
            fn(request, items, page)
        else:
            warn(
                "`construct_wagtail_userbar` hook functions should accept a `page` argument in third position -"
                f" {fn.__module__}.{fn.__name__} needs to be updated",
                category=RemovedInWagtail80Warning,
            )
            fn(request, items)


class Userbar(Component):
    template_name = "wagtailadmin/userbar/base.html"

    def __init__(self, *, object=None, position="bottom-right"):
        self.object = object
        self.position = position

    def get_context_data(self, parent_context):
        request = parent_context.get("request")
        # Render the userbar differently within the preview panel.
        in_preview_panel = getattr(request, "in_preview_panel", False)

        if not request or request.user.is_anonymous:
            language = None
        else:
            # Render the userbar using the user's preferred admin language
            userprofile = UserProfile.get_for_user(request.user)
            language = userprofile.get_preferred_language()

        with translation.override(language):
            try:
                revision_id = request.revision_id
            except AttributeError:
                revision_id = None

            if in_preview_panel:
                items = [
                    AccessibilityItem(),
                ]
            elif isinstance(self.object, Page) and self.object.pk:
                if revision_id:
                    revision = Revision.page_revisions.get(id=revision_id)
                    items = [
                        AdminItem(),
                        ExplorePageItem(revision.content_object),
                        EditPageItem(revision.content_object),
                        AccessibilityItem(),
                    ]
                else:
                    # Not a revision
                    items = [
                        AdminItem(),
                        ExplorePageItem(self.object),
                        EditPageItem(self.object),
                        AddPageItem(self.object),
                        AccessibilityItem(),
                    ]
            else:
                # Not a page.
                items = [
                    AdminItem(),
                    AccessibilityItem(),
                ]

            apply_userbar_hooks(request, items, self.object)

            # Render the items
            rendered_items = [item.render(request) for item in items]

            # Remove any unrendered items
            rendered_items = [item for item in rendered_items if item]

            if request:
                origin = f"{request.scheme}://{request.get_host()}"
            else:
                origin = get_admin_base_url() or ""

            # Render the userbar items
            return {
                "request": request,
                "origin": origin,
                "items": rendered_items,
                "position": self.position,
                "page": self.object,
                "revision_id": revision_id,
            }



from unittest import mock

from django.contrib.auth import get_user_model
from django.contrib.auth.models import Permission
from django.http import HttpRequest, HttpResponse
from django.test import TestCase
from django.urls import reverse

from wagtail.admin.views.pages.bulk_actions.page_bulk_action import PageBulkAction
from wagtail.models import Page
from wagtail.signals import post_page_move, pre_page_move
from wagtail.test.testapp.models import BusinessChild, SimplePage
from wagtail.test.utils import WagtailTestUtils


class TestBulkMove(WagtailTestUtils, TestCase):
    fixtures = ["test.json"]

    def setUp(self):
        # Find root page
        self.root_page = Page.objects.get(id=2)

        # Create three sections
        self.section_a = SimplePage(
            title="Section A", slug="section-a", content="hello"
        )
        self.root_page.add_child(instance=self.section_a)

        self.section_b = SimplePage(
            title="Section B", slug="section-b", content="hello"
        )
        self.root_page.add_child(instance=self.section_b)

        self.section_c = SimplePage(
            title="Section C", slug="section-c", content="hello"
        )
        self.root_page.add_child(instance=self.section_c)

        # Add test page A into section A
        self.test_page_a = SimplePage(
            title="Hello world!", slug="hello-world-a", content="hello"
        )
        self.section_a.add_child(instance=self.test_page_a)

        # Add test page B into section C
        self.test_page_b = SimplePage(
            title="Hello world!", slug="hello-world-b", content="hello"
        )
        self.section_c.add_child(instance=self.test_page_b)

        # Add test page B_1 into section C
        self.test_page_b_1 = SimplePage(
            title="Hello world!", slug="hello-world-b-1", content="hello"
        )
        self.section_c.add_child(instance=self.test_page_b_1)

        # Add test page A_1 into section C having same slug as test page A
        self.test_page_a_1 = SimplePage(
            title="Hello world!", slug="hello-world-a", content="hello"
        )
        self.section_c.add_child(instance=self.test_page_a_1)

        # Add unpublished page to the root with a child page
        self.unpublished_page = SimplePage(
            title="Unpublished", slug="unpublished", content="hello"
        )
        sub_page = SimplePage(title="Sub Page", slug="sub-page", content="child")
        self.root_page.add_child(instance=self.unpublished_page)
        self.unpublished_page.add_child(instance=sub_page)

        # unpublish pages last (used to validate the edit only permission)
        self.unpublished_page.unpublish()
        sub_page.unpublish()

        self.pages_to_be_moved = [self.test_page_b, self.test_page_b_1]

        self.url = (
            reverse(
                "wagtail_bulk_action",
                args=(
                    "wagtailcore",
                    "page",
                    "move",
                ),
            )
            + f"?id={self.test_page_b.id}&id={self.test_page_b_1.id}"
        )

        # Login
        self.user = self.login()

    def test_bulk_move(self):
        response = self.client.get(self.url)
        self.assertEqual(response.status_code, 200)

        html = response.content.decode()

        self.assertInHTML("<p>Are you sure you want to move these pages?</p>", html)

        self.assertInHTML(
            '<li><a href="{edit_page_url}" target="_blank" rel="noreferrer">Hello world! (simple page)</a></li>'.format(
                edit_page_url=reverse(
                    "wagtailadmin_pages:edit", args=[self.test_page_b.id]
                ),
            ),
            html,
        )

    def test_bulk_move_bad_permissions(self):
        # Remove privileges from user
        self.user.is_superuser = False
        self.user.user_permissions.add(
            Permission.objects.get(
                content_type__app_label="wagtailadmin", codename="access_admin"
            )
        )
        self.user.save()

        # Get move page
        response = self.client.get(self.url)

        self.assertEqual(response.status_code, 200)

        html = response.content.decode()

        self.assertInHTML("<p>You don't have permission to move these pages</p>", html)

        for child_page in self.pages_to_be_moved:
            self.assertInHTML(f"<li>{child_page.title}</li>", html)

        self.assertTagInHTML(
            f"""<form action="{self.url}" method="POST"></form>""",
            html,
            count=0,
        )

    def test_user_without_bulk_delete_permission_can_move(self):
        # to verify that a user without bulk delete permission is able to move a page with a child page

        self.client.logout()
        user = get_user_model().objects.get(email="siteeditor@example.com")
        self.login(user)

        # ensure the bulk_delete is not applicable to this user
        can_bulk_delete = self.test_page_b.permissions_for_user(user).can_delete()
        self.assertFalse(can_bulk_delete)

        response = self.client.get(
            reverse(
                "wagtail_bulk_action",
                args=(
                    "wagtailcore",
                    "page",
                    "move",
                ),
            )
            + f"?id={self.unpublished_page.id}"
        )

        self.assertEqual(response.status_code, 200)

    def test_bulk_move_destination_not_allowed(self):
        page = BusinessChild(title="Section no child", slug="section-no-child")
        self.root_page.add_child(instance=page)

        response = self.client.post(self.url, {"chooser": page.id})

        html = response.content.decode()

        self.assertInHTML(
            f"<p>The following pages cannot be moved to {page.title}</p>", html
        )

        for child_page in self.pages_to_be_moved:
            self.assertInHTML(
                '<li><a href="{edit_page_url}" target="_blank" rel="noreferrer">{page_title}</a></li>'.format(
                    edit_page_url=reverse(
                        "wagtailadmin_pages:edit", args=[child_page.id]
                    ),
                    page_title=child_page.title,
                ),
                html,
            )

    def test_bulk_move_slug_already_taken(self):
        temp_page_1 = SimplePage(
            title="Hello world!", slug="hello-world-b", content="hello"
        )
        temp_page_2 = SimplePage(
            title="Hello world!", slug="hello-world-b-1", content="hello"
        )
        self.section_b.add_child(instance=temp_page_1)
        self.section_b.add_child(instance=temp_page_2)

        response = self.client.post(self.url, {"chooser": self.section_b.id})

        html = response.content.decode()

        self.assertInHTML(
            "<p>The following pages cannot be moved due to duplicate slugs</p>", html
        )

        for child_page in self.pages_to_be_moved:
            self.assertInHTML(
                '<li><a href="{edit_page_url}" target="_blank" rel="noreferrer">{page_title}</a></li>'.format(
                    edit_page_url=reverse(
                        "wagtailadmin_pages:edit", args=[child_page.id]
                    ),
                    page_title=child_page.title,
                ),
                html,
            )

    def test_bulk_move_triggers_signals(self):
        # Connect a mock signal handler to pre_page_move and post_page_move signals
        pre_moved_handler = mock.MagicMock()
        post_moved_handler = mock.MagicMock()

        pre_page_move.connect(pre_moved_handler)
        post_page_move.connect(post_moved_handler)

        # Post to view confirm move page
        try:
            self.client.post(self.url, {"chooser": self.section_b.id})
        finally:
            # Disconnect mock handler to prevent cross-test pollution
            pre_page_move.disconnect(pre_moved_handler)
            post_page_move.disconnect(post_moved_handler)

        # Check that the pre_page_move signals were fired
        self.assertTrue(
            pre_moved_handler.mock_calls[0].called_with(
                sender=self.test_page_b.specific_class,
                instance=self.test_page_b,
                parent_page_before=self.section_c,
                parent_page_after=self.section_b,
                url_path_before="/home/section-c/hello-world-b/",
                url_path_after="/home/section-b/hello-world-b/",
            )
        )
        self.assertTrue(
            pre_moved_handler.mock_calls[1].called_with(
                sender=self.test_page_b_1.specific_class,
                instance=self.test_page_b_1,
                parent_page_before=self.section_c,
                parent_page_after=self.section_b,
                url_path_before="/home/section-c/hello-world-b-1/",
                url_path_after="/home/section-b/hello-world-b-1/",
            )
        )

        # Check that the post_page_move signals were fired
        self.assertTrue(
            post_moved_handler.mock_calls[0].called_with(
                sender=self.test_page_b.specific_class,
                instance=self.test_page_b,
                parent_page_before=self.section_c,
                parent_page_after=self.section_b,
                url_path_before="/home/section-c/hello-world-b/",
                url_path_after="/home/section-b/hello-world-b/",
            )
        )
        self.assertTrue(
            post_moved_handler.mock_calls[1].called_with(
                sender=self.test_page_b_1.specific_class,
                instance=self.test_page_b_1,
                parent_page_before=self.section_c,
                parent_page_after=self.section_b,
                url_path_before="/home/section-c/hello-world-b-1/",
                url_path_after="/home/section-b/hello-world-b-1/",
            )
        )

    def test_before_bulk_move_hook(self):
        def hook_func(request, action_type, pages, action_class_instance):
            self.assertEqual(action_type, "move")
            self.assertIsInstance(request, HttpRequest)
            self.assertIsInstance(action_class_instance, PageBulkAction)
            for i, page in enumerate(pages):
                self.assertEqual(page.id, self.pages_to_be_moved[i].id)

            return HttpResponse("Overridden!")

        with self.register_hook("before_bulk_action", hook_func):
            response = self.client.post(self.url, {"chooser": self.section_b.id})

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, b"Overridden!")

        self.assertEqual(
            Page.objects.get(id=self.test_page_b.id).get_parent().id, self.section_c.id
        )
        self.assertEqual(
            Page.objects.get(id=self.test_page_b_1.id).get_parent().id,
            self.section_c.id,
        )

    def test_after_bulk_move_hook(self):
        def hook_func(request, action_type, pages, action_class_instance):
            self.assertEqual(action_type, "move")
            self.assertIsInstance(request, HttpRequest)
            self.assertIsInstance(action_class_instance, PageBulkAction)
            for i, page in enumerate(pages):
                self.assertEqual(page.id, self.pages_to_be_moved[i].id)

            return HttpResponse("Overridden!")

        with self.register_hook("after_bulk_action", hook_func):
            response = self.client.post(self.url, {"chooser": self.section_b.id})

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, b"Overridden!")

        # pages should be moved
        self.assertEqual(
            Page.objects.get(id=self.test_page_b.id).get_parent().id, self.section_b.id
        )
        self.assertEqual(
            Page.objects.get(id=self.test_page_b_1.id).get_parent().id,
            self.section_b.id,
        )

    def test_bulk_move_with_restricted_parent_page_types(self):
        board_meetings = Page.objects.get(
            url_path="/home/events/businessy-events/board-meetings/"
        )
        url = (
            reverse(
                "wagtail_bulk_action",
                args=(
                    "wagtailcore",
                    "page",
                    "move",
                ),
            )
            + f"?id={board_meetings.id}"
        )
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)

        html = response.content.decode()

        self.assertInHTML("<p>Are you sure you want to move these pages?</p>", html)

        self.assertInHTML(
            '<li><a href="{edit_page_url}" target="_blank" rel="noreferrer">Board meetings</a></li>'.format(
                edit_page_url=reverse(
                    "wagtailadmin_pages:edit", args=[board_meetings.id]
                ),
            ),
            html,
        )



from django import forms
from django.core.exceptions import ValidationError
from django.template.loader import render_to_string
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _

from wagtail.admin.compare import BlockComparison, StructBlockComparison
from wagtail.admin.telepath import register
from wagtail.blocks import BooleanBlock, CharBlock, ChooserBlock, StructBlock
from wagtail.blocks.struct_block import StructBlockAdapter, StructBlockValidationError
from wagtail.images.models import AbstractImage

from .shortcuts import get_rendition_or_not_found


class ImageChooserBlock(ChooserBlock):
    @cached_property
    def target_model(self):
        from wagtail.images import get_image_model

        return get_image_model()

    @cached_property
    def widget(self):
        from wagtail.images.widgets import AdminImageChooser

        return AdminImageChooser()

    def render_basic(self, value, context=None):
        if value:
            return get_rendition_or_not_found(value, "original").img_tag()
        else:
            return ""

    def get_comparison_class(self):
        return ImageChooserBlockComparison

    class Meta:
        icon = "image"


class ImageChooserBlockComparison(BlockComparison):
    def htmlvalue(self, val):
        return render_to_string(
            "wagtailimages/widgets/compare.html",
            {
                "image_a": val,
                "image_b": val,
            },
        )

    def htmldiff(self):
        return render_to_string(
            "wagtailimages/widgets/compare.html",
            {
                "image_a": self.val_a,
                "image_b": self.val_b,
            },
        )


class ImageBlock(StructBlock):
    """
    An usage of ImageChooserBlock with support for alt text.
    For backward compatibility, this block overrides necessary methods to change
    the StructValue to be an Image model instance, making it compatible in
    places where ImageChooserBlock was used.
    """

    image = ImageChooserBlock(required=True)
    decorative = BooleanBlock(
        default=False, required=False, label=_("Image is decorative")
    )
    alt_text = CharBlock(required=False, label=_("Alt text"))

    def __init__(self, required=True, **kwargs):
        super().__init__(
            [
                ("image", ImageChooserBlock(required=required)),
                (
                    "decorative",
                    BooleanBlock(
                        default=False, required=False, label=_("Image is decorative")
                    ),
                ),
                ("alt_text", CharBlock(required=False, label=_("Alt text"))),
            ],
            **kwargs,
        )

    def deconstruct(self):
        """
        For typical StructBlock subclasses, it makes sense for the deconstructed block object to be a basic StructBlock
        with the child blocks passed to the constructor (because this is largely functionally identical to the
        subclass, and avoids embedding a reference to a potentially-volatile custom class in migrations).

        This is not the case for ImageBlock, as it overrides enough of StructBlock's behaviour that a basic StructBlock
        is not a suitable substitute - and also has an incompatible constructor signature (as we don't want to support
        passing child blocks to it).

        Therefore, we opt out of the standard StructBlock deconstruction behaviour here, and always
        deconstruct an ImageBlock as an ImageBlock.
        """
        return ("wagtail.images.blocks.ImageBlock", [], self._constructor_kwargs)

    def deconstruct_with_lookup(self, lookup):
        return self.deconstruct()

    @classmethod
    def construct_from_lookup(cls, lookup, *args, **kwargs):
        return cls(**kwargs)

    def get_searchable_content(self, value):
        if not self.search_index or not value:
            return []

        return self.child_blocks["alt_text"].get_searchable_content(
            value.contextual_alt_text
        )

    def _struct_value_to_image(self, struct_value):
        image = struct_value.get("image")
        decorative = struct_value.get("decorative")
        if image:
            # If the image is decorative, set alt_text to an empty string
            image.contextual_alt_text = (
                "" if decorative else struct_value.get("alt_text")
            )
            image.decorative = decorative
        return image

    def _image_to_struct_value(self, image):
        return {
            "image": image,
            "alt_text": image and image.contextual_alt_text,
            "decorative": image and image.decorative,
        }

    def to_python(self, value):
        # For backward compatibility with ImageChooserBlock
        if value is None or isinstance(value, int):
            image = self.child_blocks["image"].to_python(value)
            struct_value = {
                "image": image,
                "decorative": False,
                "alt_text": (image.default_alt_text if image else ""),
            }
        else:
            struct_value = super().to_python(value)
        return self._struct_value_to_image(struct_value)

    def bulk_to_python(self, values):
        values = list(values)

        if values and all(value is None or isinstance(value, int) for value in values):
            # `values` looks like a list of image IDs and/or None values (as we might encounter
            # if an ImageChooserBlock has been changed to an ImageBlock with no data migration)
            image_values = self.child_blocks["image"].bulk_to_python(values)

            struct_values = [
                {
                    "image": image,
                    "decorative": False,
                    "alt_text": (image.default_alt_text if image else ""),
                }
                for image in image_values
            ]

        else:
            # Treat `values` as the standard ImageBlock representation - a (possibly empty) list of
            # dicts containing `image`, `decorative` and `alt_text` keys to be handled by the
            # StructBlock superclass
            struct_values = super().bulk_to_python(values)

        return [
            self._struct_value_to_image(struct_value) for struct_value in struct_values
        ]

    def value_from_datadict(self, data, files, prefix):
        struct_value = super().value_from_datadict(data, files, prefix)
        return self._struct_value_to_image(struct_value)

    def clean(self, value):
        try:
            self.child_blocks["image"].clean(value)
        except ValidationError as e:
            raise StructBlockValidationError(
                block_errors={"image": e},
            )

        if value and not value.contextual_alt_text and not value.decorative:
            raise StructBlockValidationError(
                block_errors={
                    "alt_text": ValidationError(
                        _(
                            "Please add some alt text for your image or mark it as decorative"
                        )
                    )
                }
            )

        return value

    def normalize(self, value):
        if value is None or isinstance(value, AbstractImage):
            return value
        else:
            struct_value = super().normalize(value)
            return self._struct_value_to_image(struct_value)

    def get_form_context(self, value, prefix="", errors=None):
        dict_value = {
            "image": value,
            "alt_text": value and value.contextual_alt_text,
            "decorative": value and value.decorative,
        }
        context = super().get_form_context(dict_value, prefix=prefix, errors=errors)
        context["suggested_alt_text"] = value
        return context

    def get_form_state(self, value):
        return {
            "image": self.child_blocks["image"].get_form_state(value),
            "alt_text": value and value.contextual_alt_text,
            "decorative": value and value.decorative,
        }

    def get_prep_value(self, value):
        return {
            "image": self.child_blocks["image"].get_prep_value(value),
            "alt_text": value and value.contextual_alt_text,
            "decorative": value and value.decorative,
        }

    def extract_references(self, value):
        return self.child_blocks["image"].extract_references(value)

    def get_comparison_class(self):
        return ImageBlockComparison

    def get_api_representation(self, value, context=None):
        return super().get_api_representation(
            self._image_to_struct_value(value), context=context
        )

    def render_basic(self, value, context=None):
        return self.child_blocks["image"].render_basic(value, context=context)

    def get_block_by_content_path(self, value, path_elements):
        if path_elements:
            return super().get_block_by_content_path(
                self._image_to_struct_value(value), path_elements
            )
        else:
            return self.bind(value)

    class Meta:
        icon = "image"
        template = "wagtailimages/widgets/image.html"


class ImageBlockAdapter(StructBlockAdapter):
    js_constructor = "wagtail.images.blocks.ImageBlock"

    @cached_property
    def media(self):
        structblock_media = super().media
        return forms.Media(
            js=structblock_media._js + ["wagtailimages/js/image-block.js"],
            css=structblock_media._css,
        )


register(ImageBlockAdapter(), ImageBlock)


class ImageBlockComparison(StructBlockComparison):
    def __init__(self, block, exists_a, exists_b, val_a, val_b):
        super().__init__(
            block,
            exists_a,
            exists_b,
            block._image_to_struct_value(val_a),
            block._image_to_struct_value(val_b),
        )

    def htmlvalue(self, val):
        if isinstance(val, AbstractImage):
            return super().htmlvalue(self.block._image_to_struct_value(val))
        else:
            return super().htmlvalue(val)



from datetime import date
from io import StringIO

from django.core import management

from wagtail.search.query import MATCH_ALL
from wagtail.search.tests.test_backends import BackendTests
from wagtail.test.search import models


class ElasticsearchCommonSearchBackendTests(BackendTests):
    def test_search_with_spaces_only(self):
        # Search for some space characters and hope it doesn't crash
        results = self.backend.search("   ", models.Book)

        # Queries are lazily evaluated, force it to run
        list(results)

        # Didn't crash, yay!

    def test_filter_with_unsupported_lookup_type(self):
        """
        Not all lookup types are supported by the Elasticsearch backends
        """
        from wagtail.search.backends.base import FilterError

        with self.assertRaises(FilterError):
            list(
                self.backend.search(
                    "Hello", models.Book.objects.filter(title__iregex="h(ea)llo")
                )
            )

    def test_partial_search(self):
        results = self.backend.autocomplete("Java", models.Book)

        self.assertUnsortedListEqual(
            [r.title for r in results],
            ["JavaScript: The Definitive Guide", "JavaScript: The good parts"],
        )

    def test_disabled_partial_search(self):
        results = self.backend.search("Java", models.Book)

        self.assertUnsortedListEqual([r.title for r in results], [])

    def test_disabled_partial_search_with_whole_term(self):
        # Making sure that there isn't a different reason why the above test
        # returned no results
        results = self.backend.search("JavaScript", models.Book)

        self.assertUnsortedListEqual(
            [r.title for r in results],
            ["JavaScript: The Definitive Guide", "JavaScript: The good parts"],
        )

    def test_child_partial_search(self):
        # Note: Expands to "Westeros". Which is in a field on Novel.setting
        results = self.backend.autocomplete("Wes", models.Book)

        self.assertUnsortedListEqual(
            [r.title for r in results],
            ["A Game of Thrones", "A Storm of Swords", "A Clash of Kings"],
        )

    def test_ascii_folding(self):
        book = models.Book.objects.create(
            title="Ĥéllø", publication_date=date(2017, 10, 19), number_of_pages=1
        )

        index = self.backend.get_index_for_model(models.Book)
        index.add_item(book)
        index.refresh()

        results = self.backend.autocomplete("Hello", models.Book)

        self.assertUnsortedListEqual([r.title for r in results], ["Ĥéllø"])

    def test_query_analyser(self):
        # This is testing that fields that use edgengram_analyzer as their index analyser do not
        # have it also as their query analyser
        results = self.backend.search("JavaScript", models.Book)
        self.assertUnsortedListEqual(
            [r.title for r in results],
            ["JavaScript: The Definitive Guide", "JavaScript: The good parts"],
        )

        # Even though they both start with "Java", this should not match the "JavaScript" books
        results = self.backend.search("JavaBeans", models.Book)
        self.assertSetEqual({r.title for r in results}, set())

    def test_search_with_hyphen(self):
        """
        This tests that punctuation characters are treated the same
        way in both indexing and querying.

        See: https://github.com/wagtail/wagtail/issues/937
        """
        book = models.Book.objects.create(
            title="Harry Potter and the Half-Blood Prince",
            publication_date=date(2009, 7, 15),
            number_of_pages=607,
        )

        index = self.backend.get_index_for_model(models.Book)
        index.add_item(book)
        index.refresh()

        results = self.backend.search("Half-Blood", models.Book)
        self.assertUnsortedListEqual(
            [r.title for r in results],
            [
                "Harry Potter and the Half-Blood Prince",
            ],
        )

    def test_search_with_numeric_term(self):
        book = models.Book.objects.create(
            title="Harry Potter and the 31337 Goblets of Fire",
            publication_date=date(2009, 7, 15),
            number_of_pages=607,
        )

        index = self.backend.get_index_for_model(models.Book)
        index.add_item(book)
        index.refresh()

        results = self.backend.search("31337", models.Book)
        self.assertUnsortedListEqual(
            [r.title for r in results],
            [
                "Harry Potter and the 31337 Goblets of Fire",
            ],
        )

        results = self.backend.autocomplete("313", models.Book)
        self.assertUnsortedListEqual(
            [r.title for r in results],
            [
                "Harry Potter and the 31337 Goblets of Fire",
            ],
        )

        results = self.backend.search("31337 goblets", models.Book)
        self.assertUnsortedListEqual(
            [r.title for r in results],
            [
                "Harry Potter and the 31337 Goblets of Fire",
            ],
        )

        results = self.backend.autocomplete("31337 gob", models.Book)
        self.assertUnsortedListEqual(
            [r.title for r in results],
            [
                "Harry Potter and the 31337 Goblets of Fire",
            ],
        )

    def test_and_operator_with_single_field(self):
        # Testing for bug #1859
        results = self.backend.search(
            "JavaScript", models.Book, operator="and", fields=["title"]
        )
        self.assertUnsortedListEqual(
            [r.title for r in results],
            ["JavaScript: The Definitive Guide", "JavaScript: The good parts"],
        )

    def test_update_index_command_schema_only(self):
        management.call_command(
            "update_index",
            backend_name=self.backend_name,
            schema_only=True,
            stdout=StringIO(),
        )

        # This should not give any results
        results = self.backend.search(MATCH_ALL, models.Book)
        self.assertSetEqual(set(results), set())

    def test_more_than_ten_results(self):
        # #3431 reported that Elasticsearch only sends back 10 results if the results set is not sliced
        results = self.backend.search(MATCH_ALL, models.Book)

        self.assertEqual(len(results), 14)

    def test_more_than_one_hundred_results(self):
        # Tests that fetching more than 100 results uses the scroll API
        books = []
        for i in range(150):
            books.append(
                models.Book.objects.create(
                    title=f"Book {i}",
                    publication_date=date(2017, 10, 21),
                    number_of_pages=i,
                )
            )

        index = self.backend.get_index_for_model(models.Book)
        index.add_items(models.Book, books)
        index.refresh()

        results = self.backend.search(MATCH_ALL, models.Book)
        self.assertEqual(len(results), 164)

    def test_slice_more_than_one_hundred_results(self):
        books = []
        for i in range(150):
            books.append(
                models.Book.objects.create(
                    title=f"Book {i}",
                    publication_date=date(2017, 10, 21),
                    number_of_pages=i,
                )
            )

        index = self.backend.get_index_for_model(models.Book)
        index.add_items(models.Book, books)
        index.refresh()

        results = self.backend.search(MATCH_ALL, models.Book)[10:120]
        self.assertEqual(len(results), 110)

    def test_slice_to_next_page(self):
        # ES scroll API doesn't support offset. The implementation has an optimisation
        # which will skip the first page if the first result is on the second page
        books = []
        for i in range(150):
            books.append(
                models.Book.objects.create(
                    title=f"Book {i}",
                    publication_date=date(2017, 10, 21),
                    number_of_pages=i,
                )
            )

        index = self.backend.get_index_for_model(models.Book)
        index.add_items(models.Book, books)
        index.refresh()

        results = self.backend.search(MATCH_ALL, models.Book)[110:]
        self.assertEqual(len(results), 54)

    def test_cannot_filter_on_date_parts_other_than_year(self):
        # Filtering by date not supported, should throw a FilterError
        from wagtail.search.backends.base import FilterError

        in_jan = models.Book.objects.filter(publication_date__month=1)
        with self.assertRaises(FilterError):
            self.backend.search(MATCH_ALL, in_jan)



from django.contrib.admin.utils import quote
from django.contrib.auth.models import Permission
from django.http import HttpRequest, HttpResponse
from django.test import TestCase
from django.urls import reverse
from django.utils.http import urlencode

from wagtail.snippets.bulk_actions.delete import DeleteBulkAction
from wagtail.test.testapp.models import (
    Advert,
    FullFeaturedSnippet,
    VariousOnDeleteModel,
)
from wagtail.test.utils import WagtailTestUtils


class TestSnippetDeleteView(WagtailTestUtils, TestCase):
    def setUp(self):
        self.snippet_model = FullFeaturedSnippet

        # create a set of test snippets
        self.test_snippets = [
            self.snippet_model.objects.create(
                text=f"Title-{i}",
            )
            for i in range(1, 6)
        ]

        self.user = self.login()
        self.url = (
            reverse(
                "wagtail_bulk_action",
                args=(
                    self.snippet_model._meta.app_label,
                    self.snippet_model._meta.model_name,
                    "delete",
                ),
            )
            + "?"
        )

    def get_url(self, items=()):
        items = items or self.test_snippets
        return self.url + "&".join(f"id={item.pk}" for item in items)

    def test_simple(self):
        response = self.client.get(self.get_url())
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(
            response, "wagtailsnippets/bulk_actions/confirm_bulk_delete.html"
        )
        self.assertTemplateUsed(response, "wagtailadmin/shared/header.html")
        self.assertEqual(response.context["header_icon"], "cog")
        self.assertContains(response, "icon icon-cog", count=1)

    def test_get_single_delete(self):
        item = self.test_snippets[0]
        response = self.client.get(self.get_url(items=(item,)))
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(
            response, "wagtailsnippets/bulk_actions/confirm_bulk_delete.html"
        )
        self.assertTemplateUsed(response, "wagtailadmin/shared/header.html")
        self.assertEqual(response.context["header_icon"], "cog")
        self.assertContains(response, "icon icon-cog", count=1)
        self.assertContains(
            response,
            "<title>Delete full-featured snippet - Title-1 - Wagtail</title>",
            html=True,
        )
        self.assertContains(
            response,
            reverse(
                self.snippet_model.snippet_viewset.get_url_name("usage"),
                args=(quote(item.pk),),
            ),
        )
        self.assertContains(
            response,
            "This full-featured snippet is referenced 0 times.",
        )

    def test_bulk_delete(self):
        response = self.client.post(self.get_url())

        # Should redirect back to index
        self.assertEqual(response.status_code, 302)

        # Check that the users were deleted
        for snippet in self.test_snippets:
            self.assertFalse(self.snippet_model.objects.filter(pk=snippet.pk).exists())

    def test_delete_with_limited_permissions(self):
        self.user.is_superuser = False
        self.user.user_permissions.add(
            Permission.objects.get(
                content_type__app_label="wagtailadmin", codename="access_admin"
            )
        )
        self.user.save()

        response = self.client.get(self.get_url())
        self.assertEqual(response.status_code, 200)

        html = response.content.decode()
        self.assertInHTML(
            "<p>You don't have permission to delete these full-featured snippets</p>",
            html,
        )

        for snippet in self.test_snippets:
            self.assertInHTML(f"<li>{snippet.text}</li>", html)

        response = self.client.post(self.get_url())
        # User should be redirected back to the index
        self.assertEqual(response.status_code, 302)

        # Snippets should not be deleted
        for snippet in self.test_snippets:
            self.assertTrue(self.snippet_model.objects.filter(pk=snippet.pk).exists())

    def test_before_bulk_action_hook_get(self):
        with self.register_hook(
            "before_bulk_action", lambda *args: HttpResponse("Overridden!")
        ):
            response = self.client.get(self.get_url())

        self.assertEqual(response.status_code, 200)

        # The hook was not called
        self.assertNotEqual(response.content, b"Overridden!")

        # The instances were not deleted
        self.assertQuerySetEqual(
            self.snippet_model.objects.filter(
                pk__in=[snippet.pk for snippet in self.test_snippets]
            ),
            self.test_snippets,
            ordered=False,
        )

    def test_before_bulk_action_hook_post(self):
        def hook_func(request, action_type, instances, action_class_instance):
            self.assertIsInstance(request, HttpRequest)
            self.assertEqual(action_type, "delete")
            self.assertEqual(set(instances), set(self.test_snippets))
            self.assertIsInstance(action_class_instance, DeleteBulkAction)
            return HttpResponse("Overridden!")

        with self.register_hook("before_bulk_action", hook_func):
            response = self.client.post(self.get_url())

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, b"Overridden!")

        # Request intercepted before the snippets were deleted
        self.assertQuerySetEqual(
            self.snippet_model.objects.filter(
                pk__in=[snippet.pk for snippet in self.test_snippets]
            ),
            self.test_snippets,
            ordered=False,
        )

    def test_after_bulk_action_hook(self):
        def hook_func(request, action_type, instances, action_class_instance):
            self.assertIsInstance(request, HttpRequest)
            self.assertEqual(action_type, "delete")
            self.assertEqual(set(instances), set(self.test_snippets))
            self.assertIsInstance(action_class_instance, DeleteBulkAction)
            return HttpResponse("Overridden!")

        with self.register_hook("after_bulk_action", hook_func):
            response = self.client.post(self.get_url())

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, b"Overridden!")

        # Request intercepted after the snippets were deleted
        self.assertFalse(
            self.snippet_model.objects.filter(
                pk__in=[snippet.pk for snippet in self.test_snippets]
            ).exists()
        )

    # Also tests that the {before,after}_delete_snippet hooks are called.
    # These hooks have existed since before bulk actions were introduced,
    # so we need to make sure they still work.

    def test_before_delete_snippet_hook_get(self):
        def hook_func(request, instances):
            self.assertIsInstance(request, HttpRequest)
            self.assertEqual(set(instances), set(self.test_snippets))
            return HttpResponse("Overridden!")

        with self.register_hook("before_delete_snippet", hook_func):
            response = self.client.get(self.get_url())

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, b"Overridden!")

        # Request intercepted before the snippets were deleted
        self.assertQuerySetEqual(
            self.snippet_model.objects.filter(
                pk__in=[snippet.pk for snippet in self.test_snippets]
            ),
            self.test_snippets,
            ordered=False,
        )

    def test_before_delete_snippet_hook_post(self):
        def hook_func(request, instances):
            self.assertIsInstance(request, HttpRequest)
            self.assertEqual(set(instances), set(self.test_snippets))
            return HttpResponse("Overridden!")

        with self.register_hook("before_delete_snippet", hook_func):
            response = self.client.post(self.get_url())

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, b"Overridden!")

        # Request intercepted before the snippets were deleted
        self.assertQuerySetEqual(
            self.snippet_model.objects.filter(
                pk__in=[snippet.pk for snippet in self.test_snippets]
            ),
            self.test_snippets,
            ordered=False,
        )

    def test_after_delete_snippet_hook(self):
        def hook_func(request, instances):
            self.assertIsInstance(request, HttpRequest)
            self.assertEqual(set(instances), set(self.test_snippets))
            return HttpResponse("Overridden!")

        with self.register_hook("after_delete_snippet", hook_func):
            response = self.client.post(self.get_url())

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content, b"Overridden!")

        # Request intercepted after the snippets were deleted
        self.assertFalse(
            self.snippet_model.objects.filter(
                pk__in=[snippet.pk for snippet in self.test_snippets]
            ).exists()
        )


class TestProtectedBulkDeleteView(WagtailTestUtils, TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.snippet_model = Advert
        cls.test_snippets = [
            cls.snippet_model.objects.create(text=f"Title-{i}") for i in range(1, 6)
        ]
        cls.url = reverse(
            "wagtail_bulk_action",
            args=(
                cls.snippet_model._meta.app_label,
                cls.snippet_model._meta.model_name,
                "delete",
            ),
        )
        cls.query_params = {
            "next": reverse("wagtailsnippets_tests_advert:list"),
            "id": [item.pk for item in cls.test_snippets],
        }
        cls.url += "?" + urlencode(cls.query_params, doseq=True)

    def setUp(self):
        self.user = self.login()

    def test_delete_get_with_protected_reference(self):
        protected = self.test_snippets[0]
        with self.captureOnCommitCallbacks(execute=True):
            VariousOnDeleteModel.objects.create(
                text="Undeletable",
                on_delete_protect=protected,
            )
        response = self.client.get(self.url)
        self.assertEqual(response.status_code, 200)
        soup = self.get_soup(response.content)
        main = soup.select_one("main")
        usage_link = main.find(
            "a",
            href=reverse(
                "wagtailsnippets_tests_advert:usage",
                args=[quote(protected.pk)],
            )
            + "?describe_on_delete=1",
        )
        self.assertIsNotNone(usage_link)
        self.assertEqual(usage_link.text.strip(), "This advert is referenced 1 time.")
        self.assertContains(
            response,
            "One or more references to this advert prevent it from being deleted.",
        )
        submit_button = main.select_one("form button[type=submit]")
        self.assertIsNone(submit_button)
        back_button = main.find("a", href=reverse("wagtailsnippets_tests_advert:list"))
        self.assertIsNotNone(back_button)
        self.assertEqual(back_button.text.strip(), "Go back")

    def test_delete_post_with_protected_reference(self):
        protected = self.test_snippets[0]
        with self.captureOnCommitCallbacks(execute=True):
            VariousOnDeleteModel.objects.create(
                text="Undeletable",
                on_delete_protect=protected,
            )
        response = self.client.post(self.url)

        # Should throw a PermissionDenied error and redirect to the dashboard
        self.assertEqual(response.status_code, 302)
        self.assertRedirects(response, reverse("wagtailadmin_home"))
        self.assertEqual(
            response.context["message"],
            "Sorry, you do not have permission to access this area.",
        )

        # Check that the snippet is still here
        self.assertTrue(Advert.objects.filter(pk=protected.pk).exists())
